from functools import cache
import numpy as np 
from scipy.stats import pearsonr
from sklearn.neighbors import NearestNeighbors

from tqdm import tqdm
from pqdm.processes import pqdm
from velocyto.estimation import colDeltaCorpartial, colDeltaCor


def compute_probability(i, d_i, gene_mtx, indices, n_cells, T):
    exp_corr_sum = 0
    row_probs = np.zeros(n_cells)
    
    for j in indices[i]:
        r_ij = gene_mtx[i] - gene_mtx[j]
        corr, _ = pearsonr(r_ij, d_i)
        if np.isnan(corr):
            corr = 1
        exp_corr = np.exp(corr / T)
        exp_corr_sum += exp_corr
        row_probs[j] = exp_corr

    if exp_corr_sum != 0:
        row_probs /= exp_corr_sum

    return np.array(row_probs)

## CellOracle uses adapted Velocyto code
## This function is coded exactly as described in CellOracle paper
def estimate_transition_probabilities(adata, delta_X, embedding=None, n_neighbors=200, 
random_neighbors=False, annot=None, T=0.05, n_jobs=1):

    n_cells, n_genes = adata.shape
    delta_X = np.array(delta_X)
    gene_mtx = adata.layers['imputed_count']

    if n_neighbors is None:

        P = np.ones((n_cells, n_cells))

        corr = colDeltaCor(
            np.ascontiguousarray(gene_mtx.T), 
            np.ascontiguousarray(delta_X.T), 
            threads=n_jobs
            )
    
    else:

        P = np.zeros((n_cells, n_cells))

        n_neighbors = min(n_cells, n_neighbors)
        
        if random_neighbors == 'even':

            cts = np.unique(adata.obs[annot])
            ct_dict = {ct: np.where(adata.obs[annot] == ct)[0] for ct in cts}
            cells_per_ct = round(n_neighbors / len(cts))

            indices = []

            for i in range(n_cells):
                i_indices = []

                for ct, ct_cells in ct_dict.items():

                    sample = np.random.choice(ct_cells[ct_cells != i], size=cells_per_ct, replace=False)
                    i_indices.extend(sample)

                i_indices = np.array(i_indices)
                P[i, i_indices] = 1
                indices.append(i_indices)

            indices = np.array(indices)

        elif random_neighbors:
            
            indices = []
            cells = np.arange(n_cells)
            for i in range(n_cells):
                i_indices = np.random.choice(np.delete(cells, i), size=n_neighbors, replace=False)
                P[i, i_indices] = 1
                indices.append(i_indices)
            
            indices = np.array(indices)

        else: 

            nn = NearestNeighbors(n_neighbors=n_neighbors, n_jobs=n_jobs)
            nn.fit(embedding)
            _, indices = nn.kneighbors(embedding)

            rows = np.repeat(np.arange(n_cells), n_neighbors)
            cols = indices.flatten()
            P[rows, cols] = 1
        
        gene_mtx = gene_mtx.astype('float64')
        corr = colDeltaCorpartial(
            np.ascontiguousarray(gene_mtx.T), 
            np.ascontiguousarray(delta_X.T), 
            indices, threads=n_jobs
        )

    corr = np.nan_to_num(corr, nan=1)

    np.fill_diagonal(P, 0)
    P *= np.exp(corr / T)   
    P /= P.sum(1)[:, None]

    # args = [[i, delta_X[i], gene_mtx, indices, n_cells, T] for i in range(n_cells)]
    # results = pqdm(
    #     args,
    #     compute_probability,
    #     n_jobs=n_jobs,
    #     argument_type='args',
    #     tqdm_class=tqdm,
    #     desc='Estimating cell transition probabilities',
    # )

    # for i, row_probs in enumerate(results):
    #     P[i] = np.array(row_probs)

    return P


def project_probabilities(P, embedding, normalize=True):
    if normalize: 
        embed_dim = embedding.shape[1]

        embedding_T = embedding.T # shape (m, n_cells)
        unitary_vectors = embedding_T[:, None, :] - embedding_T[:, :, None]  # shape (m, n_cells, n_cells)
        unitary_vectors = unitary_vectors.astype(np.float64)
        
        # Normalize the difference vectors (L2 norm)
        with np.errstate(divide='ignore', invalid='ignore'):
            norms = np.linalg.norm(unitary_vectors, ord=2, axis=0)  # shape (n_cells, n_cells)
            unitary_vectors /= norms
            for m in range(embed_dim):
                np.fill_diagonal(unitary_vectors[m, ...], 0)   
        
        delta_embedding = (P * unitary_vectors).sum(2)  # shape (m, n_cells)
        delta_embedding = delta_embedding.T
        
        return delta_embedding
    
    else:
        embed_diffs = embedding[np.newaxis, :, :] - embedding[:, np.newaxis, :]
        
        # masked = embed_diffs * P[:, :, np.newaxis]
        # V_simulated = np.sum(masked, axis=1)
        V_simulated = np.einsum('ij,ijk->ik', P, embed_diffs)
        
        return V_simulated
    




import numpy as np
import warnings

from SpaceTravLR.models.pixel_attention import NicheAttentionNetwork

warnings.filterwarnings("ignore", category=DeprecationWarning)
from pysal.model.spreg import OLS
from abc import ABC, abstractmethod
import copy
from tqdm import tqdm 
import enlighten
from numba import jit
import torch.nn.functional as F
import torch
# import commot as ct
import torch.nn as nn
from torch.nn.utils.parametrizations import weight_norm
from torch.utils.data import DataLoader, TensorDataset, random_split
from torchvision.transforms import Normalize
from .spatial_map import xyc2spatial
from .vit_blocks import ViT

from ..tools.utils import set_seed, seed_worker, deprecated
from ..tools.data import SpaceOracleDataset
from ..tools.network import GeneRegulatoryNetwork, DayThreeRegulatoryNetwork, SurveyRegulatoryNetwork
from ..tools.network import expand_paired_interactions
set_seed(42)


if torch.backends.mps.is_available():
    device = torch.device("mps")
elif torch.cuda.is_available():
    device = torch.device("cuda")
else:
    device = torch.device("cpu")

norm = Normalize(0, 1)


class AbstractEstimator(ABC):
    
    def __init__(self):
        pass
        
    @abstractmethod
    def fit(self):
        pass
    
    @abstractmethod
    def get_betas(self):
        pass

    def predict_y(self):
        raise NotImplementedError
    
    def _training_loop(self):
        raise NotImplementedError
    
    def _validation_loop(self):
        raise NotImplementedError
    

    
class LeastSquaredEstimator(AbstractEstimator):
    
    def fit(self, X, y):
        ols_model = OLS(y=y, x=X)
        self.betas = ols_model.betas
        self.pvals = np.array(ols_model.t_stat)[:, 1]
    
    def get_betas(self):
        return self.betas

class ClusterLeastSquaredEstimator(LeastSquaredEstimator):
    
    def fit(self, X, y, clusters):
        self.beta_dict = {}
        self.pval_dict = {}
        self.clusters = clusters
        for cluster_label in np.unique(clusters):
            ols_model = OLS(y=y[clusters==cluster_label], x=X[clusters==cluster_label])
            self.beta_dict[cluster_label] = ols_model.betas
            self.pval_dict[cluster_label] = np.array(ols_model.t_stat)[:, 1]

    def get_betas(self, cluster_label):
        return self.beta_dict[self.betas]


class SimpleCNN(nn.Module):
    @deprecated('Please use ViT instead.')
    def __init__(self, nbetas, spatial_dim=64,in_channels=1, init=0.1):
        set_seed(42)
        super().__init__()
        self.dim = nbetas
        # self.betas = torch.tensor(betas.astype(np.float32)).to(device)
        
        self.conv_layers = nn.Sequential(
            weight_norm(nn.Conv2d(in_channels, 32, kernel_size=3, padding='same')),
            nn.PReLU(init=init),
            nn.MaxPool2d(kernel_size=2, stride=2),
            
            weight_norm(nn.Conv2d(32, 64, kernel_size=3, padding='same')),
            nn.PReLU(init=init),
            nn.MaxPool2d(kernel_size=2, stride=2),
            
            weight_norm(nn.Conv2d(64, 256, kernel_size=3, padding='same')),
            nn.PReLU(init=init),
            nn.MaxPool2d(kernel_size=2, stride=2),
            
            nn.AdaptiveAvgPool2d(1),
            nn.Flatten()
        )
        self.fc_layers = nn.Sequential(
            nn.Linear(256, 128),
            nn.PReLU(init=init),
            
            nn.Linear(128, 64),
            nn.PReLU(init=init),
            
            nn.Linear(64, 16),
            nn.PReLU(init=init),
            nn.Dropout(0.2),
            nn.Linear(16, self.dim)
        )

    def forward(self, spatial_map):
        spatial_features = self.conv_layers(spatial_map)
        betas = self.fc_layers(spatial_features)
        return betas


class VisionEstimator(AbstractEstimator):
    def __init__(self, adata, target_gene, annot='rctd_cluster', 
            grn=None, regulators=None, df_ligrec=None, layer='imputed_count'):
        assert target_gene in adata.var_names
        assert layer in adata.layers

        self.adata = adata
        self.annot = annot
        self.target_gene = target_gene
        
        if regulators == None:
            self.grn = DayThreeRegulatoryNetwork() # CellOracle GRN
            self.regulators = self.grn.get_cluster_regulators(self.adata, self.target_gene)
        else:
            self.regulators = regulators


        if df_ligrec is None:
            # df_ligrec = ct.pp.ligand_receptor_database(
            #     database='CellChat', 
            #     species='mouse', 
            #     signaling_type="Secreted Signaling"
            # )
            
            # df_ligrec.columns = ['ligand', 'receptor', 'pathway', 'signaling']
            df_ligrec = get_cellchat_db('mouse')

        self.lr = expand_paired_interactions(df_ligrec)
        self.lr = self.lr[self.lr.ligand.isin(adata.var_names) & (self.lr.receptor.isin(adata.var_names))]

        self.lr['pairs'] = self.lr.ligand.values + '$' + self.lr.receptor.values
        self.lr = self.lr.drop_duplicates(subset='pairs', keep='first')
        self.ligands = list(self.lr.ligand.values)
        self.receptors = list(self.lr.receptor.values)
        self.n_clusters = len(self.adata.obs[annot].unique())
        
        self.modulators = self.regulators + list(self.lr['pairs'])
        self.modulators_genes = list(np.unique(self.regulators+self.ligands+self.receptors))

        self.layer = layer
        self.model = None
        self.losses = []

        self.test_mode = False
        
        # assert len(self.regulators) > 0, f'No regulators found for target gene {self.target_gene}.'

    def predict_y(self, model, betas, batch_labels, inputs_x):

        assert inputs_x.shape[1] == len(self.regulators) == model.dim-1
        assert betas.shape[1] == len(self.regulators)+1 == len(model.betas) == len(self.regulators)+1, f'{betas.shape} {model.betas.shape}'

        y_pred = betas[:, 0]
         
        for w in range(model.dim-1):
            y_pred += betas[:, w+1]*inputs_x[:, w]

        return y_pred

    
    def _mask_betas(self, betas, batch_labels):
        regulator_dict = self.grn.regulator_dict
        relevant_tfs = [regulator_dict[label.item()] for label in batch_labels]
        mask = torch.stack(relevant_tfs).to(device)
        return betas * mask

    def _training_loop(self, model, dataloader, criterion, optimizer, cluster_grn=False):
        model.train()
        total_loss = 0

        for batch_spatial, batch_x, batch_y, batch_labels in dataloader:
            
            optimizer.zero_grad()
            betas = model(batch_spatial.to(device), batch_labels.to(device))

            if cluster_grn:
                betas = self._mask_betas(betas, batch_labels)
            
            outputs = self.predict_y(model, betas, batch_labels, inputs_x=batch_x.to(device))

            loss = criterion(outputs.squeeze(), batch_y.to(device).squeeze())

            # loss += 1e-3*torch.mean(
            #     (betas - torch.zeros_like(betas).float().to(device))**2)

            loss.backward()
            optimizer.step()
            total_loss += loss.item()
                    
        return total_loss / len(dataloader)
    
    @torch.no_grad()
    def _validation_loop(self, model, dataloader, criterion, cluster_grn = False):
        model.eval()
        total_loss = 0
        for batch_spatial, batch_x, batch_y, batch_labels in dataloader:
            betas = model(batch_spatial.to(device), batch_labels.to(device))
            
            if cluster_grn:
                betas = self._mask_betas(betas, batch_labels)
            
            outputs = self.predict_y(model, betas, batch_labels, inputs_x=batch_x.to(device), anchors=None)
            loss = criterion(outputs.squeeze(), batch_y.to(device).squeeze())
            total_loss += loss.item()
        
        return total_loss / len(dataloader)
    
    def _estimate_baseline(self, dataloader, beta_init):
        total_linear_err = 0
        torch.manual_seed(42)
        for _, batch_x, batch_y, _ in dataloader:
            _x = batch_x.cpu().numpy()
            _y = batch_y.cpu().numpy()
            # batch_coefs = batch_coefs.cpu().numpy().reshape(-1, )
            
            ols_pred = beta_init[0]
            for w in range(len(beta_init)-1):
                ols_pred += _x[:, w]*beta_init[w+1]

            # ols_pred = 0
            # for w in range(len(beta_init)-1):
            #     ols_pred += _x[:, w]*batch_coefs[w]
                
            ols_err = np.mean((_y - ols_pred)**2)
            
            total_linear_err += ols_err
            
        return total_linear_err / len(dataloader)
        
    @staticmethod
    def _build_dataloaders_from_adata(adata, target_gene, regulators, batch_size=32, 
    mode='train', rotate_maps=True, annot='rctd_cluster', layer='imputed_count', spatial_dim=64, test_size=0.2):

        assert mode in ['train', 'train_test']
        set_seed(42)

        xy = adata.obsm['spatial']
        labels = np.array(adata.obs[annot])
    
        g = torch.Generator()
        g.manual_seed(42)
        
        params = {
            'batch_size': batch_size,
            'worker_init_fn': seed_worker,
            'generator': g,
            'pin_memory': False,
            'num_workers': 0,
            'drop_last': True,
        }
        
        dataset = SpaceOracleDataset(
            adata.copy(), 
            target_gene=target_gene, 
            regulators=regulators, 
            annot=annot, 
            layer=layer,
            spatial_dim=spatial_dim,
            rotate_maps=rotate_maps
        )

        if mode == 'train':
            train_dataloader = DataLoader(dataset, shuffle=True, **params)
            valid_dataloader = DataLoader(dataset, shuffle=False, **params)
            return train_dataloader, valid_dataloader
        
        if mode == 'train_test':
            split = int((1-test_size)*len(dataset))
            generator = torch.Generator().manual_seed(42)
            train_dataset, valid_dataset = random_split(
                dataset, [split, len(dataset)-split], generator=generator)
            train_dataloader = DataLoader(train_dataset, shuffle=True, **params)
            valid_dataloader = DataLoader(valid_dataset, shuffle=False, **params)

            return train_dataloader, valid_dataloader
        
        
    @torch.no_grad()
    def get_betas(self, xy=None, spatial_maps=None, labels=None, spatial_dim=None, layer=None):

        assert xy is not None or spatial_maps is not None

        spatial_dim = self.spatial_dim if spatial_dim is None else spatial_dim
        
        if spatial_maps is None:
            # spatial_maps = norm(
            #     torch.from_numpy(
            #         xyc2spatial(
            #             xy[:, 0], xy[:, 1], 
            #             labels, spatial_dim, spatial_dim, 
            #             disable_tqdm=False
            #         )
            #     ).float()
            # )
            spatial_maps = torch.from_numpy(
                xyc2spatial(
                    xy[:, 0], xy[:, 1], 
                    labels, spatial_dim, spatial_dim, 
                    disable_tqdm=False
                ).float()
            )   
        else:
            spatial_maps = torch.from_numpy(spatial_maps)

        dataset = TensorDataset(
            spatial_maps.float(), 
            torch.from_numpy(labels).long()
        )   

        g = torch.Generator()
        g.manual_seed(42)
        
        params = {
            'batch_size': 1024,
            'worker_init_fn': seed_worker,
            'generator': g
        }
        
        infer_dataloader = DataLoader(dataset, shuffle=False, **params)

        beta_list = []
            
        for batch_spatial, batch_labels in infer_dataloader:
            betas = self.model(batch_spatial.to(device), batch_labels.to(device))
            beta_list.extend(betas.cpu().numpy())
        
        return np.array(beta_list)

        # if self.cluster_grn:
        #     beta_stack = []
            
        #     for batch_spatial, batch_labels in infer_dataloader:
        #         betas = self.model(batch_spatial.to(device), batch_labels.to(device))
        #         betas = self._mask_betas(betas, batch_labels)
        #         beta_stack.append(betas)
        
        #     beta_stack = torch.cat(beta_stack)
        #     return beta_stack.cpu().numpy()

        # else:
        #     beta_list = []
            
        #     for batch_spatial, batch_labels in infer_dataloader:
        #         betas = self.model(batch_spatial.to(device), batch_labels.to(device))
        #         beta_list.extend(betas.cpu().numpy())
            
        #     return np.array(beta_list)

# class GeoCNNEstimatorV2(VisionEstimator):
#     def _build_cnn(
#         self, 
#         adata,
#         annot,
#         spatial_dim,
#         mode, 
#         max_epochs,
#         batch_size, 
#         learning_rate,
#         rotate_maps
#         ):


#         train_dataloader, valid_dataloader = self._build_dataloaders_from_adata(
#                 adata, self.target_gene, self.regulators, 
#                 mode=mode, rotate_maps=rotate_maps, batch_size=batch_size,
#                 annot=annot, spatial_dim=spatial_dim)
           
#         model = BetaModel(self.beta_init, in_channels=self.n_clusters)
#         criterion = nn.MSELoss(reduction='mean')
#         optimizer = torch.optim.Adam(model.parameters(), lr=learning_rate)
        
#         model.to(device)
#         # model = torch.compile(model)
        
#         losses = []
#         best_model = copy.deepcopy(model)
#         best_score = np.inf
#         best_iter = 0
    
#         baseline_loss = self._estimate_baseline(valid_dataloader, self.beta_init)
            
#         with tqdm(range(max_epochs)) as pbar:
#             for epoch in pbar:
#                 training_loss = self._training_loop(model, train_dataloader, criterion, optimizer)
#                 validation_loss = self._validation_loop(model, valid_dataloader, criterion)
                
#                 losses.append(validation_loss)

#                 pbar.set_description(f'[{device.type}] MSE: {np.mean(losses):.4f} | Baseline: {baseline_loss:.4f}')
            
#                 if validation_loss < best_score:
#                     best_score = validation_loss
#                     best_model = copy.deepcopy(model)
#                     best_iter = epoch
            
#         best_model.eval()
        
#         print(f'Best model at {best_iter}/{max_epochs}')
        
#         return best_model, losses
        
#     def fit(
#         self,
#         annot,
#         init_betas='ols', 
#         max_epochs=100, 
#         learning_rate=0.001, 
#         spatial_dim=64,
#         batch_size=32, 
#         mode='train',
#         rotate_maps=True
#         ):
        
        
#         assert init_betas in ['ones', 'ols']
        
#         self.spatial_dim = spatial_dim  

#         adata = self.adata.copy()

#         if init_betas == 'ones':
#             beta_init = torch.ones(len(self.regulators)+1)
        
#         elif init_betas == 'ols':
#             X = adata.to_df()[self.regulators].values
#             y = adata.to_df()[[self.target_gene]].values
#             ols = LeastSquaredEstimator()
#             ols.fit(X, y)
#             beta_init = ols.get_betas()
            
#         self.beta_init = np.array(beta_init).reshape(-1, )
        
#         try:
#             model, losses = self._build_cnn(
#                 adata,
#                 annot,
#                 spatial_dim=spatial_dim, 
#                 mode=mode,
#                 max_epochs=max_epochs,
#                 batch_size=batch_size,
#                 learning_rate=learning_rate,
#                 rotate_maps=rotate_maps
#             ) 
            
#             self.model = model  
#             self.losses = losses
            
        
#         except KeyboardInterrupt:
#             print('Training interrupted...')
#             pass
        
        

class ViTEstimatorV2(VisionEstimator):
    def _build_model(
        self,
        adata,
        annot,
        spatial_dim,
        mode,
        max_epochs,
        batch_size,
        learning_rate,
        rotate_maps,
        regularize,
        cluster_grn,
        n_patches=2, n_blocks=4, hidden_d=14, n_heads=2,
        pbar=None
        ):

        train_dataloader, valid_dataloader = self._build_dataloaders_from_adata(
                adata, self.target_gene, self.regulators, 
                mode=mode, rotate_maps=rotate_maps, batch_size=batch_size,
                annot=annot, layer=self.layer, spatial_dim=spatial_dim)
        
        # self.train_dataloader = train_dataloader
        # self.valid_dataloader = valid_dataloader

        model = ViT(
            self.beta_init, 
            in_channels=self.n_clusters, 
            spatial_dim=spatial_dim, 
            n_patches=n_patches, 
            n_blocks=n_blocks, 
            hidden_d=hidden_d, 
            n_heads=n_heads
        )
        
        criterion = nn.MSELoss(reduction='mean')
        optimizer = torch.optim.Adam(model.parameters(), lr=learning_rate)
        
        model.to(device)
        # model = torch.compile(model)
        
        losses = []
        best_model = copy.deepcopy(model)
        best_score = np.inf
        best_iter = 0
    
        baseline_loss = self._estimate_baseline(valid_dataloader, self.beta_init)
        _prefix = f'[{self.target_gene} / {len(self.regulators)}]'

        if pbar is None:
            _manager = enlighten.get_manager()
            pbar = _manager.counter(
                total=max_epochs, 
                desc=f'{_prefix} <> MSE: ... | Baseline: {baseline_loss:.4f}', 
                unit='epochs'
            )
            pbar.refresh()
            
        for epoch in range(max_epochs):
            training_loss = self._training_loop(model, train_dataloader, criterion, 
                                optimizer, cluster_grn=cluster_grn, regularize=regularize)
            validation_loss = self._validation_loop(model, valid_dataloader, 
                                criterion, cluster_grn=cluster_grn)
            
            losses.append(validation_loss)

            if validation_loss < best_score:
                best_score = validation_loss
                best_model = copy.deepcopy(model)
                best_iter = epoch
            
            pbar.desc = f'{_prefix} <> MSE: {np.mean(losses):.4f}'
            pbar.update()
            
        best_model.eval()
        
        return best_model, losses

    
    def fit(
        self,
        annot,
        init_betas='ols', 
        max_epochs=100, 
        learning_rate=2e-4, 
        spatial_dim=64,
        batch_size=32, 
        mode='train',
        regularize=False,
        cluster_grn=False,
        rotate_maps=True,
        n_patches=2, n_blocks=2, hidden_d=16, n_heads=2,
        pbar=None
        ):
        
        assert annot in self.adata.obs.columns
        assert init_betas in ['ones', 'ols', 'co']

        self.spatial_dim = spatial_dim  
        self.regularize = regularize
        self.cluster_grn = cluster_grn
        self.rotate_maps = rotate_maps
        self.init_betas = init_betas
        self.annot = annot

        adata = self.adata.copy()

        if init_betas == 'ones':
            beta_init = torch.ones(len(self.regulators)+1)
        
        elif init_betas == 'ols':
            X = adata.to_df(layer=self.layer)[self.regulators].values
            y = adata.to_df(layer=self.layer)[[self.target_gene]].values
            ols = LeastSquaredEstimator()
            ols.fit(X, y)
            beta_init = ols.get_betas()

        elif init_betas == 'co':
            co_coefs = self.grn.get_regulators_with_pvalues(
                adata, self.target_gene).groupby('source').mean()
            co_coefs = co_coefs.loc[self.regulators]
            beta_init = np.array(co_coefs.values).reshape(-1, )
            beta_init = np.concatenate([[1], beta_init], axis=0) 
            
        self.beta_init = np.array(beta_init).reshape(-1, )

        assert len(self.beta_init) == len(self.regulators)+1
        
        try:
            model, losses = self._build_model(
                adata,
                annot,
                spatial_dim=spatial_dim, 
                mode=mode,
                max_epochs=max_epochs,
                batch_size=batch_size,
                learning_rate=learning_rate,
                rotate_maps=rotate_maps,
                regularize=regularize,
                cluster_grn=cluster_grn,
                n_patches=n_patches, 
                n_blocks=n_blocks, 
                hidden_d=hidden_d, 
                n_heads=n_heads,
                pbar=pbar
            )
            
            self.model = model  
            self.losses = losses
            
        
        except KeyboardInterrupt:
            print('Training interrupted...')
            pass


    def export(self):
        self.model.eval()
        # self.model.cpu()
        return self.model, self.regulators, self.target_gene


class PixelAttention(VisionEstimator):
    def _build_model(
        self,
        adata,
        annot,
        spatial_dim,
        mode,
        layer,
        max_epochs,
        batch_size,
        learning_rate,
        rotate_maps,
        cluster_grn,
        regularize,
        pbar=None
        ):

        train_dataloader, valid_dataloader = self._build_dataloaders_from_adata(
            adata, self.target_gene, self.regulators, 
            mode=mode, rotate_maps=rotate_maps, 
            batch_size=batch_size, annot=annot, 
            layer=layer,
            spatial_dim=spatial_dim
        )

        model = NicheAttentionNetwork(
            betas=self.beta_init,
            in_channels=self.n_clusters,
            spatial_dim=spatial_dim,
        )


        model.to(device)

        criterion = nn.MSELoss(reduction='mean')
        optimizer = torch.optim.Adam(model.parameters(), lr=learning_rate)

        losses = []
        best_model = copy.deepcopy(model)
        best_score = np.inf
        best_iter = 0
    
        baseline_loss = self._estimate_baseline(valid_dataloader, self.beta_init)
        _prefix = f'[{self.target_gene} / {len(self.regulators)}]'

        if pbar is None:
            _manager = enlighten.get_manager()
            pbar = _manager.counter(
                total=max_epochs, 
                desc=f'{_prefix} <> MSE: ... | Baseline: {baseline_loss:.4f}', 
                unit='epochs'
            )
            pbar.refresh()
            
        for epoch in range(max_epochs):
            training_loss = self._training_loop(
                model, train_dataloader, criterion, optimizer, 
                cluster_grn=cluster_grn, regularize=regularize)
            validation_loss = self._validation_loop(
                model, valid_dataloader, criterion, 
                cluster_grn=cluster_grn)
            
            losses.append(validation_loss)

            if validation_loss < best_score:
                best_score = validation_loss
                best_model = copy.deepcopy(model)
                best_iter = epoch
            
            pbar.desc = f'{_prefix} <> MSE: {np.mean(losses):.4g}'
            pbar.update()
            
        best_model.eval()
        
        return best_model, losses


    def fit(
        self,
        annot,
        init_betas='zeros', 
        max_epochs=10, 
        learning_rate=2e-4, 
        spatial_dim=64,
        batch_size=32, 
        mode='train',
        regularize=False,
        rotate_maps=True,
        cluster_grn=False,
        pbar=None
        ):
        
        assert annot in self.adata.obs.columns
        assert init_betas in ['ones', 'ols', 'zeros', 'co']

        self.spatial_dim = spatial_dim  
        self.regularize = regularize
        self.rotate_maps = rotate_maps
        self.init_betas = init_betas
        self.annot = annot
        self.cluster_grn = cluster_grn

        adata = self.adata.copy()

        if init_betas == 'ones':
            beta_init = torch.ones(len(self.regulators)+1)
        
        elif init_betas == 'ols':
            X = adata.to_df()[self.regulators].values
            y = adata.to_df()[[self.target_gene]].values
            ols = LeastSquaredEstimator()
            ols.fit(X, y)
            beta_init = ols.get_betas()

        elif init_betas == 'co':
            co_coefs = self.grn.get_regulators_with_pvalues(
                adata, self.target_gene).groupby('source').mean()
            co_coefs = co_coefs.loc[self.regulators]
            beta_init = np.array(co_coefs.values).reshape(-1, )
            beta_init = np.concatenate([[1], beta_init], axis=0) 

        elif init_betas == 'zeros':
            beta_init = torch.zeros(len(self.regulators)+1)
            
        self.beta_init = np.array(beta_init).reshape(-1, )

        assert len(self.beta_init) == len(self.regulators)+1
        
        try:
            model, losses = self._build_model(
                adata,
                annot,
                spatial_dim=spatial_dim, 
                mode=mode,
                layer=self.layer,
                cluster_grn=cluster_grn,
                max_epochs=max_epochs,
                batch_size=batch_size,
                learning_rate=learning_rate,
                rotate_maps=rotate_maps,
                regularize=regularize,
                pbar=pbar
            )
            
            self.model = model  
            self.losses = losses
            
        
        except KeyboardInterrupt:
            print('Training interrupted...')
            pass


    def export(self):
        self.model.eval()
        # self.model.cpu()
        return self.model, self.regulators, self.target_gene



from pysal.model.spreg import OLS
from scipy.stats import t
from sklearn.metrics import r2_score
from torch.utils.data import DataLoader, TensorDataset, random_split
import pandas as pd
import numpy as np
import copy
import enlighten
import torch
import pyro
import torch.nn as nn
import copy
import os
import pickle
from tqdm import tqdm
from sklearn.linear_model import BayesianRidge
from SpaceTravLR.models.spatial_map import xyc2spatial_fast
from SpaceTravLR.models.estimators import LeastSquaredEstimator, VisionEstimator
from SpaceTravLR.models.vit_blocks import ViT
from SpaceTravLR.tools.data import LigRecDataset
from .pixel_attention import NicheAttentionNetwork
from .bayesian_linear import BayesianRegression
from ..tools.utils import gaussian_kernel_2d, set_seed, seed_worker

set_seed(42)

device = torch.device(
    "mps" if torch.backends.mps.is_available() 
    else "cuda" if torch.cuda.is_available() 
    else "cpu"
)

available_cores = os.cpu_count()

pyro.clear_param_store()

cmn = lambda x, y: len(np.intersect1d(x, y))

class ProbabilisticPixelAttention(VisionEstimator):

    def _build_model(
        self,
        adata,
        annot,
        spatial_dim,
        mode,
        layer,
        max_epochs,
        batch_size,
        learning_rate,
        rotate_maps,
        pbar=None
        ):

        train_dataloader, valid_dataloader = self._build_dataloaders_from_adata(
            adata, self.target_gene, self.regulators, 
            mode=mode, rotate_maps=rotate_maps, 
            batch_size=batch_size, annot=annot, 
            layer=layer,
            spatial_dim=spatial_dim
        )

        model = NicheAttentionNetwork(
            n_regulators=len(self.regulators),
            in_channels=self.n_clusters,
            spatial_dim=spatial_dim,
        )


        model.to(device)

        criterion = nn.MSELoss(reduction='mean')
        optimizer = torch.optim.Adam(model.parameters(), lr=learning_rate)

        losses = []
        best_model = copy.deepcopy(model)
        best_score = np.inf
        best_iter = 0
    
        # baseline_loss = self._estimate_baseline(valid_dataloader, self.beta_init)
        _prefix = f'[{self.target_gene} / {len(self.regulators)}]'

        if pbar is None:
            _manager = enlighten.get_manager()
            pbar = _manager.counter(
                total=max_epochs, 
                desc=f'{_prefix} <> MSE: ...', 
                unit='epochs'
            )
            pbar.refresh()

        for epoch in range(max_epochs):
            training_loss = self._training_loop(
                model, train_dataloader, criterion, optimizer)
            validation_loss = self._validation_loop(
                model, valid_dataloader, criterion)
            
            losses.append(validation_loss)

            if validation_loss < best_score:
                best_score = validation_loss
                best_model = copy.deepcopy(model)
                best_iter = epoch
            
            pbar.desc = f'{_prefix} <> MSE: {np.mean(losses):.4g}'
            pbar.update()
            
        best_model.eval()
        
        return best_model, losses
    
    def predict_y(self, model, betas, batch_labels, inputs_x, anchors=None):

        assert inputs_x.shape[1] == len(self.regulators) == model.dim-1
        assert betas.shape[1] == len(self.regulators)+1

        if anchors is None:
            anchors = np.stack(
                [self.beta_dists[label].mean(0) for label in batch_labels.cpu().numpy()], 
                axis=0
            )
        
        anchors = torch.from_numpy(anchors).float().to(device)

        y_pred = anchors[:, 0]*betas[:, 0]
         
        for w in range(model.dim-1):
            y_pred += anchors[:, w+1]*betas[:, w+1]*inputs_x[:, w]

        return y_pred
    
    
    def _training_loop(self, model, dataloader, criterion, optimizer):
        model.train()
        total_loss = 0

        for batch_spatial, batch_x, batch_y, batch_labels in dataloader:
            
            optimizer.zero_grad()
            betas = model(batch_spatial.to(device), batch_labels.to(device))

            anchors = np.stack(
                [self.beta_dists[label].mean(0) for label in batch_labels.cpu().numpy()], 
                axis=0
            )
            
            outputs = self.predict_y(model, betas, batch_labels, inputs_x=batch_x.to(device), anchors=anchors)

            loss = criterion(outputs.squeeze(), batch_y.to(device).squeeze())
            loss += 1e-3*((betas.mean(0) - torch.from_numpy(anchors).float().mean(0).to(device))**2).sum()
            
            loss.backward()
            optimizer.step()
            total_loss += loss.item()
                    
        return total_loss / len(dataloader)
    

    # To test if values are significant in a Bayesian model, we can use the posterior distributions of the parameters.
    # A common approach is to compute the credible intervals (CIs) for the parameters of interest.
    # If the credible interval does not include zero, we can consider the effect to be significant.
    # So a marginal posterior distribution for a given IV that does not include 0 in the 95% HDI 
    # just shows that 95% of the most likely parameter values based on the data do not include zero. 
    
    def test_significance(self, betas, alpha=0.05):
        lower_bound = np.percentile(betas, 100 * (alpha / 2), axis=0)
        upper_bound = np.percentile(betas, 100 * (1 - alpha / 2), axis=0)
        significant = (lower_bound > 0) | (upper_bound < 0)
        
        return significant



    def fit(
        self,
        annot,
        max_epochs=10, 
        learning_rate=2e-4, 
        spatial_dim=64,
        batch_size=32, 
        alpha=0.05,
        num_samples=1000,
        mode='train_test',
        rotate_maps=True,
        parallel=True,
        cache=False,
        pbar=None
        ):
        
        assert annot in self.adata.obs.columns

        self.spatial_dim = spatial_dim  
        self.rotate_maps = rotate_maps
        self.annot = annot

        adata = self.adata
        beta_dists_file = f"/tmp/{self.target_gene}_beta_dists.pkl"
        cache_exists = os.path.exists(beta_dists_file)
            
        X = torch.from_numpy(adata.to_df(layer=self.layer)[self.regulators].values).float()
        y = torch.from_numpy(adata.to_df(layer=self.layer)[self.target_gene].values).float()
        cluster_labels = torch.from_numpy(np.array(adata.obs[self.annot])).long()
        
        if cache_exists and cache:
            with open(beta_dists_file, 'rb') as f:
                self.beta_dists = pickle.load(f)

        else:
            self.beta_model = BayesianRegression(
                n_regulators=len(self.regulators), 
                device=torch.device('cpu') ## use cpu for better parallelization
            )

            _max_epochs = 1 if self.test_mode else 1000
            ns = 1 if self.test_mode else num_samples

            self.beta_model.fit(
                X, y, cluster_labels, 
                max_epochs=_max_epochs, learning_rate=3e-3, 
                num_samples=ns,
                parallel=parallel
            )

            self.beta_dists = {}
            for cluster in range(self.n_clusters):
                self.beta_dists[cluster] = self.beta_model.get_betas(
                    X[cluster_labels==cluster].to(self.beta_model.device), 
                    cluster=cluster, 
                    num_samples=ns
                )
        
            with open(beta_dists_file, 'wb') as f:
                pickle.dump(self.beta_dists, f)

        

        self.is_real = pd.DataFrame(
            [self.test_significance(self.beta_dists[i][:, 1:], alpha=alpha) for i in self.beta_dists.keys()], 
            columns=self.regulators
        ).T

        for c in self.is_real.columns:
            for ix, s in enumerate(self.is_real[c].values):
                if not s:
                    self.beta_dists[c][:, ix+1] = 0


        del X, y, cluster_labels

        try:
            _max_epochs = 1 if self.test_mode else max_epochs

            model, losses = self._build_model(
                adata,
                annot,
                spatial_dim=spatial_dim, 
                mode=mode,
                layer=self.layer,
                max_epochs=_max_epochs,
                batch_size=batch_size,
                learning_rate=learning_rate,
                rotate_maps=rotate_maps,
                pbar=pbar
            )
            
            self.model = model  
            self.losses = losses
            
        
        except KeyboardInterrupt:
            print('Training interrupted...')


    def export(self):
        self.model.eval()

        return (
            self.model, 
            self.beta_dists, 
            self.is_real, 
            self.regulators, 
            self.target_gene
        )
    
            
    @torch.no_grad()
    def get_betas(self, xy=None, spatial_maps=None, labels=None, spatial_dim=None, beta_dists=None, layer=None):

        assert xy is not None or spatial_maps is not None
        assert beta_dists is not None or self.beta_dists is not None



        spatial_dim = self.spatial_dim if spatial_dim is None else spatial_dim
        
        if spatial_maps is None:
            spatial_maps = xyc2spatial_fast(
                xyc = np.column_stack([xy, labels]),
                m=self.spatial_dim,
                n=self.spatial_dim,
            ).astype(np.float32)
            
        
        spatial_maps = torch.from_numpy(spatial_maps)

        dataset = TensorDataset(
            spatial_maps.float(), 
            torch.from_numpy(labels).long()
        )   

        g = torch.Generator()
        g.manual_seed(42)
        
        params = {
            'batch_size': 1024,
            'worker_init_fn': seed_worker,
            'generator': g
        }
        
        infer_dataloader = DataLoader(dataset, shuffle=False, **params)

        beta_list = []
            
        for batch_spatial, batch_labels in infer_dataloader:
            betas = self.model(batch_spatial.to(device), batch_labels.to(device))
            beta_list.extend(betas.cpu().numpy())
        
        return np.array(beta_list)



class ProbabilisticPixelModulators(ProbabilisticPixelAttention):
    

    @staticmethod
    def received_ligands(xy, lig_df, radius=200):
        # gex_df = adata.to_df(layer=layer)
        ligands = lig_df.columns
        gauss_weights = [
            gaussian_kernel_2d(
                xy[i], 
                xy, 
                radius=radius) for i in range(len(lig_df)
            )
        ]
        u_ligands = list(np.unique(ligands))
        wL = lambda x: np.array([(gauss_weights[i] * lig_df[x]).mean(0) for i in range(len(lig_df))])
        weighted_ligands = [wL(x) for x in u_ligands]

        return pd.DataFrame(
            weighted_ligands, 
            index=u_ligands, 
            columns=lig_df.index
        ).T




    def _build_dataloaders_from_adata(self,
        adata, target_gene, 
        regulators, ligands, receptors,
        batch_size=32, radius=200,
        mode='train', rotate_maps=True, 
        annot='rctd_cluster', layer='imputed_count', 
        spatial_dim=64, test_size=0.2
    ):

        assert mode in ['train', 'train_test']
        set_seed(42)

        g = torch.Generator()
        g.manual_seed(42)


        self.adata.uns['received_ligands'] = self.received_ligands(
            self.adata.obsm['spatial'], 
            self.adata.to_df(layer=self.layer)[np.unique(self.ligands)], 
            radius=self.radius,
        )
        
        params = {
            'batch_size': batch_size,
            'worker_init_fn': seed_worker,
            'generator': g,
            'pin_memory': False,
            'num_workers': 0,
            'drop_last': True,
        }
        
        dataset = LigRecDataset(
            adata, 
            target_gene=target_gene, 
            regulators=regulators, 
            ligands=ligands,
            receptors=receptors,
            radius=radius,
            annot=annot, 
            layer=layer,
            spatial_dim=spatial_dim,
            rotate_maps=rotate_maps
        )

        if mode == 'train':
            train_dataloader = DataLoader(dataset, shuffle=True, **params)
            valid_dataloader = DataLoader(dataset, shuffle=False, **params)
            return train_dataloader, valid_dataloader
        
        if mode == 'train_test':
            split = int((1-test_size)*len(dataset))
            generator = torch.Generator().manual_seed(42)
            train_dataset, valid_dataset = random_split(
                dataset, [split, len(dataset)-split], generator=generator)
            train_dataloader = DataLoader(train_dataset, shuffle=True, **params)
            valid_dataloader = DataLoader(valid_dataset, shuffle=False, **params)

            return train_dataloader, valid_dataloader
        
        
    def _build_model(
        self,
        adata,
        annot,
        spatial_dim,
        mode,
        layer,
        max_epochs,
        batch_size,
        learning_rate,
        rotate_maps,
        pbar=None
        ):

        train_dataloader, valid_dataloader = self._build_dataloaders_from_adata(
            adata, self.target_gene, self.regulators, self.ligands, self.receptors,
            mode=mode, rotate_maps=rotate_maps, radius=self.radius,
            batch_size=batch_size, annot=annot, 
            layer=layer,
            spatial_dim=spatial_dim
        )


        model = NicheAttentionNetwork(
            n_regulators=len(self.regulators)+len(self.ligands),
            in_channels=self.n_clusters,
            spatial_dim=spatial_dim,
        )

        # beta_init = torch.ones(len(self.regulators)+len(self.ligands)+1)
        # model = ViT(
        #     beta_init, 
        #     in_channels=self.n_clusters, 
        #     spatial_dim=spatial_dim, 
        #     n_patches=8, 
        #     n_blocks=4, 
        #     hidden_d=16, 
        #     n_heads=2
        # )

        model.to(device)

        # print(model)

        criterion = nn.MSELoss(reduction='mean')
        optimizer = torch.optim.Adam(model.parameters(), lr=learning_rate)

        losses = []
        best_model = copy.deepcopy(model)
        best_score = np.inf
        best_iter = 0
    
        # baseline_loss = self._estimate_baseline(valid_dataloader, self.beta_init)
        _prefix = f'[{self.target_gene} / {len(self.regulators)}]'

        real_modulators = self.is_real.loc[self.is_real.sum(1) > 0].index.tolist()

        print(f'{self.target_gene} > {cmn(self.regulators, real_modulators)} regulators + {cmn(self.lr["pairs"], real_modulators)} ligand-receptor pairs')

        if pbar is None:
            _manager = enlighten.get_manager()
            pbar = _manager.counter(
                total=max_epochs, 
                desc=f'{_prefix} <> MSE: ...', 
                unit='epochs'
            )
            pbar.refresh()

        for epoch in range(max_epochs):
            training_loss = self._training_loop(
                model, train_dataloader, criterion, optimizer)
            validation_loss = self._validation_loop(
                model, valid_dataloader, criterion)
            
            losses.append(validation_loss)

            if validation_loss < best_score:
                best_score = validation_loss
                best_model = copy.deepcopy(model)
                best_iter = epoch
            
            pbar.desc = f'{_prefix} <> MSE: {np.mean(losses):.4g}'
            pbar.update()
            
        best_model.eval()
        
        return best_model, losses
    
    def predict_y(self, model, betas, batch_labels, inputs_x, anchors=None):

        assert inputs_x.shape[1] == len(self.regulators)+len(self.ligands) == model.dim-1
        assert betas.shape[1] == len(self.regulators)+len(self.ligands)+1

        if anchors is None:
            anchors = np.stack(
                [self.beta_dists[label].mean(0) for label in batch_labels.cpu().numpy()], 
                axis=0
            )
        
        anchors = torch.from_numpy(anchors).float().to(device)

        y_pred = anchors[:, 0]*betas[:, 0]
         
        for w in range(model.dim-1):
            y_pred += anchors[:, w+1]*betas[:, w+1]*inputs_x[:, w]

        return y_pred
    
    
    def _training_loop(self, model, dataloader, criterion, optimizer):
        model.train()
        total_loss = 0

        for batch_spatial, batch_x, batch_y, batch_labels in dataloader:
            
            optimizer.zero_grad()
            betas = model(batch_spatial.to(device), batch_labels.to(device))


            anchors = np.stack(
                [self.beta_dists[label].mean(0) for label in batch_labels.cpu().numpy()], 
                axis=0
            )
            
            outputs = self.predict_y(model, betas, batch_labels, inputs_x=batch_x.to(device), anchors=anchors)

            loss = criterion(outputs.squeeze(), batch_y.to(device).squeeze())
            loss += 1e-3*((betas.mean(0) - torch.from_numpy(anchors).float().mean(0).to(device))**2).sum()
            
            loss.backward()
            optimizer.step()
            total_loss += loss.item()
                    
        return total_loss / len(dataloader)
    

    # To test if values are significant in a Bayesian model, we can use the posterior distributions of the parameters.
    # A common approach is to compute the credible intervals (CIs) for the parameters of interest.
    # If the credible interval does not include zero, we can consider the effect to be significant.
    # So a marginal posterior distribution for a given IV that does not include 0 in the 95% HDI 
    # just shows that 95% of the most likely parameter values based on the data do not include zero. 
    
    # def test_significance(self, betas, alpha=0.05, effect_size_threshold=0.1):
    #     lower_bound = np.percentile(betas, 100 * (alpha / 2), axis=0)
    #     upper_bound = np.percentile(betas, 100 * (1 - alpha / 2), axis=0)
        
    #     # Test for Type I error (false positive)
    #     significant_type1 = (lower_bound > 0) | (upper_bound < 0)
        
    #     # Test for Type II error (false negative)
    #     mean_effect = np.mean(betas, axis=0)
    #     significant_type2 = np.abs(mean_effect) > effect_size_threshold
        
    #     # Combine both tests
    #     significant = significant_type1 & significant_type2
        
    #     return significant

    def test_significance(self, betas, alpha=0.05):
        lower_bound = np.percentile(betas, 100 * (alpha / 2), axis=0)
        upper_bound = np.percentile(betas, 100 * (1 - alpha / 2), axis=0)
        significant = (lower_bound > 0) | (upper_bound < 0)
        
        # Calculate p-values
        t_statistic = np.mean(betas, axis=0) / (np.std(betas, axis=0) / np.sqrt(betas.shape[0]))
        p_values = 2 * (1 - t.cdf(np.abs(t_statistic), df=betas.shape[0] - 1))
        
        return significant, p_values



    def fit(
        self,
        annot,
        max_epochs=10, 
        learning_rate=2e-4, 
        spatial_dim=64,
        batch_size=32, 
        alpha=0.05,
        num_samples=1000,
        radius=200,
        mode='train_test',
        rotate_maps=True,
        parallel=True,
        cache=False,
        pbar=None
        ):
        
        assert annot in self.adata.obs.columns

        self.spatial_dim = spatial_dim  
        self.rotate_maps = rotate_maps
        self.annot = annot
        self.radius = radius

        adata = self.adata
        beta_dists_file = f"/tmp/{self.target_gene}_beta_dists.pkl"
        cache_exists = os.path.exists(beta_dists_file)
            
        X = torch.from_numpy(adata.to_df(layer=self.layer)[self.regulators].values).float()
        y = torch.from_numpy(adata.to_df(layer=self.layer)[self.target_gene].values).float()
        cluster_labels = torch.from_numpy(np.array(adata.obs[self.annot])).long()
        
        # Generate ligand-receptor data
        xy = np.array(adata.obsm['spatial']).copy()
        ligX = adata.to_df(layer=self.layer)[self.ligands].values
        recpX = adata.to_df(layer=self.layer)[self.receptors].values

        # Calculate weights using gaussian kernel
        lr_exp = []
        for i in range(len(xy)):
            w = gaussian_kernel_2d(xy[i], xy, radius=self.radius)

            ligand_exp = (ligX.T * w).T
            receptor_exp = recpX[i]
            lr_exp.append((ligand_exp * receptor_exp).mean(axis=0))

        lr_exp = torch.from_numpy(np.stack(lr_exp, axis=0)).float()

        X = torch.cat([X, lr_exp], dim=1)

        if cache_exists and cache:
            with open(beta_dists_file, 'rb') as f:
                self.beta_dists = pickle.load(f)

        else:
            self.beta_model = BayesianRegression(
                n_regulators=len(self.regulators)+len(self.ligands), 
                device=torch.device('cpu') ## use cpu for better parallelization
            )

            ns = 1 if self.test_mode else num_samples
            _max_epochs = 1 if self.test_mode else 3000


            self.beta_model.fit(
                X, y, cluster_labels, 
                max_epochs=_max_epochs, learning_rate=3e-3, 
                num_samples=ns,
                parallel=parallel
            )

            self.beta_dists = {}
            for cluster in range(self.n_clusters):
                self.beta_dists[cluster] = self.beta_model.get_betas(
                    X[cluster_labels==cluster].to(self.beta_model.device), 
                    cluster=cluster, 
                    num_samples=ns
                )

            # p_values = []

            # for cluster in range(self.n_clusters):
            #     # m = BayesianRidge()
            #     m = OLS(y.numpy()[cluster_labels==cluster], X.numpy()[cluster_labels==cluster])
            #     self.beta_dists[cluster] = m.betas
            #     p_values.append(m.pvals)


            with open(beta_dists_file, 'wb') as f:
                pickle.dump(self.beta_dists, f)

        

        self.is_real = pd.DataFrame(
            [self.test_significance(self.beta_dists[i][:, 1:], alpha=alpha)[0] for i in self.beta_dists.keys()], 
            columns=list(self.regulators)+list(self.lr['pairs'])
        ).T


        self.pvals = pd.DataFrame(
            [self.test_significance(self.beta_dists[i][:, 1:], alpha=alpha)[1] for i in self.beta_dists.keys()], 
            columns=list(self.regulators)+list(self.lr['pairs'])
        ).T


        for c in self.is_real.columns:
            for ix, s in enumerate(self.is_real[c].values):
                if not s:
                    self.beta_dists[c][:, ix+1] = 0


        del X, y, cluster_labels

        # return

        try:
            _max_epochs = 1 if self.test_mode else max_epochs

            model, losses = self._build_model(
                adata,
                annot,
                spatial_dim=spatial_dim, 
                mode=mode,
                layer=self.layer,
                max_epochs=_max_epochs,
                batch_size=batch_size,
                learning_rate=learning_rate,
                rotate_maps=rotate_maps,
                pbar=pbar
            )
            
            self.model = model  
            self.losses = losses
        
        except KeyboardInterrupt:
            print('Training interrupted...')


            
    @torch.no_grad()
    def get_betas(self, xy=None, spatial_maps=None, labels=None, spatial_dim=None, beta_dists=None):

        assert xy is not None or spatial_maps is not None
        assert beta_dists is not None or self.beta_dists is not None
        spatial_dim = self.spatial_dim if spatial_dim is None else spatial_dim
        
        if spatial_maps is None:
            spatial_maps = xyc2spatial_fast(
                xyc = np.column_stack([xy, labels]),
                m=self.spatial_dim,
                n=self.spatial_dim,
            ).astype(np.float32)

        if beta_dists is None:
            beta_dists = self.beta_dists
        
        spatial_maps = torch.from_numpy(spatial_maps)

        dataset = TensorDataset(
            spatial_maps.float(), 
            torch.from_numpy(labels).long()
        )   

        g = torch.Generator()
        g.manual_seed(42)
        
        params = {
            'batch_size': 1024,
            'worker_init_fn': seed_worker,
            'generator': g
        }
        
        infer_dataloader = DataLoader(dataset, shuffle=False, **params)

        beta_list = []
            
        for batch_spatial, batch_labels in infer_dataloader:
            anchors = np.stack(
                [beta_dists[label].mean(0) for label in batch_labels.cpu().numpy()], 
                axis=0
            )
            
            betas = self.model(batch_spatial.to(device), batch_labels.to(device))
            beta_list.extend(anchors*betas.cpu().numpy())
        
        return np.array(beta_list)
    

    def export(self):
        self.model.eval()

        return (
            self.model, 
            self.beta_dists, 
            self.is_real, 
            self.regulators, 
            self.target_gene
        )
    
    @property
    def betadata(self):
        betas = self.get_betas(
            spatial_maps=np.array(self.adata.obsm['spatial_maps']),
            labels=np.array(self.adata.obs[self.annot]))
        
        betas_df = pd.DataFrame(
            betas, 
            columns=['beta0']+['beta_'+i for i in self.is_real.index], 
            index=self.adata.obs.index
        )
        
        gex_df = self.adata.to_df(layer=self.layer)
        received_ligands = self.adata.uns['received_ligands']

        ## wL is the amount of ligand 'received' at each location
        ## assuming ligands and receptors expression are independent, dL/dR = 0
        ## y = b0 + b1*TF1 + b2*wL1R1 + b3*wL1R2
        ## dy/dTF1 = b1
        ## dy/dwL1 = b2[wL1*dR1/dwL1 + R1] + b3[wL1*dR2/dwL1 + R2]
        ##         = b2*R1 + b3*R2
        ## dy/dR1 = b2*[wL1 + R1*dwL1/dR1] = b2*wL1


        b_ligand = lambda x, y: betas_df[f'beta_{x}${y}']*received_ligands[x]
        b_receptor = lambda x, y: betas_df[f'beta_{x}${y}']*gex_df[y]

        ## dy/dR
        ligand_betas = pd.DataFrame(
            [b_ligand(x, y).values for x, y in zip(self.ligands, self.receptors)],
            columns=self.adata.obs.index, index=['beta_'+k for k in self.ligands]).T
        
        ## dy/dwL
        receptor_betas = pd.DataFrame(
            [b_receptor(x, y).values for x, y in zip(self.ligands, self.receptors)],
            columns=self.adata.obs.index, index=['beta_'+k for k in self.receptors]).T
        
        ## linearly combine betas for the same ligands or receptors
        ligand_betas = ligand_betas.groupby(lambda x:x, axis=1).sum()
        receptor_betas = receptor_betas.groupby(lambda x: x, axis=1).sum()

        assert not any(ligand_betas.columns.duplicated())
        assert not any(receptor_betas.columns.duplicated())
        
        xy = pd.DataFrame(self.adata.obsm['spatial'], index=self.adata.obs.index, columns=['x', 'y'])
        gex_modulators = self.regulators+self.ligands+self.receptors+[self.target_gene]
        

        """
        # Combine all relevant data into a single DataFrame
        # one row per cell
        betas_df \                                      # beta coefficients, TFs and LR-pairs
            .join(gex_df[np.unique(gex_modulators)]) \  # gene expression data for each modulator
            .join(self.adata.uns['ligand_receptor']) \  # weighted-ligands*receptor values
            .join(ligand_betas) \                       # beta_wLR * wL, 
            .join(receptor_betas) \                     # beta_wLR * R
            .join(self.adata.obs) \                     # cell type metadata
            .join(xy)                                   # spatial coordinates

        """

        
        _data = betas_df \
            .join(gex_df[np.unique(gex_modulators)]) \
            .join(self.adata.uns['ligand_receptor']) \
            .join(ligand_betas) \
            .join(receptor_betas) \
            .join(self.adata.obs) \
            .join(xy)
        
        inputs_x = _data[self.regulators+list(self.lr['pairs'])].values
        betas = _data[[f'beta_{i}' for i in self.regulators+list(self.lr['pairs'])]].values
        y_pred = _data['beta0'].values
        for i in range(inputs_x.shape[1]):
            y_pred += inputs_x[:, i] * betas[:, i]
        _data[f'target_{self.target_gene}'] = y_pred
        return _data





import numpy as np 
import pandas as pd 
import matplotlib.pyplot as plt 
import seaborn as sns

from SpaceTravLR.plotting.shift import estimate_transition_probabilities
from scipy.stats import spearmanr

def quantify_correlations(adata, annot='cell_type', use_pearson=False):
    celltypes = adata.obs[annot].unique()

    corr_mat = np.zeros((len(celltypes), len(celltypes)))

    for i, source_ct in enumerate(celltypes):
        source_idxs = adata.obs[annot] == source_ct
        
        for j, target_ct in enumerate(celltypes):
            target_idxs = adata.obs[annot] == target_ct
            
            if use_pearson:
                corrs = np.corrcoef(
                    adata.layers['imputed_count'][source_idxs], 
                    adata.layers['simulated_count'][target_idxs]
                )
            else:
                corrs = spearmanr(
                    adata.layers['imputed_count'][source_idxs], 
                    adata.layers['simulated_count'][target_idxs],
                    axis=1
                ).statistic

            x = source_idxs.sum()
            y = target_idxs.sum()
            cross_corr = corrs[:x, x:]
            orig_corr = corrs[:x, :y]
            corr_mat[i, j] = cross_corr.mean() - orig_corr.mean()
    
    heatmap_df = pd.DataFrame(corr_mat, index=celltypes, columns=celltypes)
    return heatmap_df


def plot_corr_heatmap(heatmap_df, title='Cross-correlation between imputed and simulated counts', save_path=None):

    # heatmap_df = sankey_df.pivot(index='source', columns='target', values='value').fillna(0)

    plt.figure(figsize=(15, 12))
    sns.heatmap(
        heatmap_df, 
        cmap='Spectral', 
        annot=True, 
        cbar=True, 
        center=0,
        linewidths=.5, 
        linecolor='black'
    )
    plt.title(title)
    plt.xlabel('Perturbed Cells')
    plt.ylabel('Original Cells')
    plt.xticks(rotation=45)
    plt.yticks(rotation=0)
    plt.tight_layout()

    if save_path is not None:
        plt.savefig(save_path)
    else:
        plt.show()
    


import os
import tempfile
import urllib.request
from abc import ABC, abstractmethod
from glob import glob
import anndata
import scanpy as sc
import warnings
import numpy as np
from torch.utils.data import Dataset
from ..models.spatial_map import xyc2spatial_fast
import torch
import pandas as pd
import os
import tempfile
import urllib.request

warnings.simplefilter(action='ignore', category=anndata.ImplicitModificationWarning)

tt = lambda x: torch.from_numpy(x.copy()).float()

class SpatialDataset(Dataset, ABC):

        
    def __len__(self):
        return self.adata.shape[0]
    
    @staticmethod
    def load_slideseq(path):
        assert '.h5ad' in path
        return anndata.read_h5ad(path)
    
    @staticmethod
    def load_visium(path):
        raise NotImplementedError
    
    @staticmethod
    def load_xenium(path):
        raise NotImplementedError
    
    @abstractmethod
    def __getitem__(self, index):
        pass
        
    

class SpaceOracleDataset(SpatialDataset):
    """
    returns spatial_info, tf_exp, target_ene_exp, cluster_info
    """

    def __init__(self, adata, target_gene, regulators, spatial_dim=16, 
    annot='rctd_cluster', layer='imputed_count', rotate_maps=True):

        self.adata = adata
        
        self.target_gene = target_gene
        self.regulators = regulators
        self.layer = layer
        self.spatial_dim = spatial_dim
        self.rotate_maps = rotate_maps
        
        self.X = adata.to_df(layer=layer)[self.regulators].values
        self.y = adata.to_df(layer=layer)[[self.target_gene]].values
        self.clusters = np.array(self.adata.obs[annot])
        self.n_clusters = len(np.unique(self.clusters))
        self.xy = np.array(self.adata.obsm['spatial'])

        if 'spatial_maps' in self.adata.obsm:
            self.spatial_maps = self.adata.obsm['spatial_maps']
        else:
            self.spatial_maps = xyc2spatial_fast(
                xyc = np.column_stack([self.xy, self.clusters]),
                m=self.spatial_dim,
                n=self.spatial_dim,
            ).astype(np.float32)
            
            self.adata.obsm['spatial_maps'] = self.spatial_maps

    def __getitem__(self, index):
        sp_map = self.spatial_maps[index]
        if self.rotate_maps:
            k = np.random.choice([0, 1, 2, 3])
            sp_map = np.rot90(sp_map, k=k, axes=(1, 2))
        spatial_info = torch.from_numpy(sp_map.copy()).float()
        tf_exp = torch.from_numpy(self.X[index].copy()).float()
        target_gene_exp = torch.from_numpy(self.y[index].copy()).float()
        cluster_info = torch.tensor(self.clusters[index]).long()

        assert spatial_info.shape[0] == self.n_clusters
        assert spatial_info.shape[1] == spatial_info.shape[2] == self.spatial_dim

        return spatial_info, tf_exp, target_gene_exp, cluster_info


class LigRecDataset(SpaceOracleDataset):
    def __init__(
            self, adata, target_gene, regulators, ligands, receptors, radius=200,
            spatial_dim=32, annot='rctd_cluster', layer='imputed_count', rotate_maps=True
        ):
        super().__init__(adata, target_gene, regulators, spatial_dim=spatial_dim, 
                                annot=annot, layer=layer, rotate_maps=rotate_maps)
        self.ligands = ligands
        self.receptors = receptors
        self.radius = radius

        # sq.gr.spatial_neighbors(adata, n_neighs=neighbors)

        xy = np.array(self.adata.obsm['spatial']).copy()
        ligX =  self.adata.to_df(layer=self.layer)[self.ligands].values
        recpX =  self.adata.to_df(layer=self.layer)[self.receptors].values


        # @staticmethod
        # def calculate_ligand_receptor_exp(adata, ligands, receptors, radius):
        #     xy = np.array(adata.obsm['spatial'])
        #     ligX = adata.to_df(layer='imputed_count')[ligands].values
        #     recpX = adata.to_df(layer='imputed_count')[receptors].values
            
        #     lr_exp = np.array([((ligX * gaussian_kernel_2d(
        #         c, xy, radius=radius)[:, np.newaxis])
        #         .mean(axis=0) * recpX[ix]) for ix, c in enumerate(xy)])
            
        #     return pd.DataFrame(
        #         lr_exp,
        #         columns=[i[0]+'$'+i[1] for i in zip(ligands, receptors)],
        #         index=adata.obs.index
        #     )


        received_ligands = self.adata.uns['received_ligands'][self.ligands].values
        receptor_expression = self.adata.to_df(layer=self.layer)[self.receptors].values

        self.lr_exp  = received_ligands * receptor_expression
        
        self.adata.uns['ligand_receptor'] = pd.DataFrame(
            self.lr_exp, 
            columns=[i[0]+'$'+i[1] for i in zip(self.ligands, self.receptors)], 
            index=self.adata.obs.index
        )

        assert self.lr_exp.shape[0] == self.adata.shape[0]
        assert self.lr_exp.shape[1] == len(self.ligands) == len(self.receptors)



    def _process_spatial(self, sp_map):
        if self.rotate_maps:
            k = np.random.choice([0, 1, 2, 3])
            sp_map = np.rot90(sp_map, k=k, axes=(1, 2))
        return sp_map


    def __getitem__(self, index):
        sp_map = self.spatial_maps[index]
        sp_map = self._process_spatial(sp_map)
        
        sp_map = tt(sp_map)
        tf_exp = tt(self.X[index])
        lr_exp = tt(self.lr_exp[index])
        target_gene_exp = tt(self.y[index])
        cluster_info = torch.tensor(self.clusters[index]).long()

        assert sp_map.shape[0] == self.n_clusters
        assert sp_map.shape[1] == sp_map.shape[2] == self.spatial_dim

        x_exp = torch.cat([tf_exp, lr_exp], dim=0)

        return sp_map, x_exp, target_gene_exp, cluster_info



from .parallel_estimators import *
from .pixel_attention import CellularNicheNetwork
from ..tools.network import expand_paired_interactions
from ..tools.utils import is_mouse_data

import torch
import torch.nn as nn 
# import commot as ct
import itertools


class PrefeaturizedTensorDataset(Dataset):
    def __init__(self, sp_maps, X_cell, y_cell):
        self.sp_maps = sp_maps
        self.X_cell = X_cell
        self.y_cell = y_cell

    def __len__(self):
        return len(self.X_cell)

    def __getitem__(self, idx):
        sp_map = self.sp_maps[idx, :]
        sp_map = np.expand_dims(sp_map, axis=0)

        return (
            torch.from_numpy(sp_map.copy()).float(),
            torch.from_numpy(self.X_cell[idx]).float(),
            torch.from_numpy(np.array(self.y_cell[idx])).float()
        )
      
class PrefeaturizedNicheNetwork(nn.Module):

    def __init__(self, n_modulators, spatial_dim=64**2):
        super().__init__()
        self.in_channels = 1
        self.out_channels = 1
        self.spatial_dim = spatial_dim
        self.dim = n_modulators+1

        self.mlp = nn.Sequential(
            nn.Linear(spatial_dim, 64),
            nn.PReLU(init=0.1),
            nn.Linear(64, self.dim)
        )
        
        self.output_activation = nn.Sigmoid()
        
    @staticmethod
    def predict_y(inputs_x, betas):
        return torch.matmul(
                inputs_x.unsqueeze(1), 
                betas[:, 1:].unsqueeze(2)
            ).squeeze(1).squeeze(1) + \
                betas[:, 0]

    def get_betas(self, spatial_maps):
        
        n_samples = spatial_maps.shape[0]
        betas = self.mlp(spatial_maps.reshape(n_samples, self.spatial_dim))
        betas = self.output_activation(betas) * 1.5

        return betas
    
    def forward(self, spatial_maps, inputs_x):

        betas = self.get_betas(spatial_maps)
        y_pred = self.predict_y(inputs_x, betas)
        
        return y_pred

class PrefeaturizedCellularProgramsEstimator(SpatialCellularProgramsEstimator):

    def __init__(self, adata, target_gene, layer='imputed_count', 
            radius=100, contact_distance=30, use_ligands=True,
            tf_ligand_cutoff=0.01, receptor_thresh=0.1,
            regulators=None, grn=None, colinks_path=None, scale_factor=1,
            sp_maps_key='COVET_SQRT'):

        assert sp_maps_key in adata.obsm.keys(), f'adata.obsm does not contain {sp_maps_key}'
        sp_maps = adata.obsm[sp_maps_key].reshape(adata.n_obs, -1)

        super().__init__(adata, target_gene, spatial_dim=sp_maps.shape[1], layer=layer, 
            radius=radius, contact_distance=contact_distance, use_ligands=use_ligands,
            tf_ligand_cutoff=tf_ligand_cutoff, receptor_thresh=receptor_thresh,
            regulators=regulators, grn=grn, colinks_path=colinks_path, scale_factor=scale_factor)
        
        self.sp_maps_key = sp_maps_key


    def init_data(self):
        '''
        Initialize the data for the estimator, without processing spatial maps
        '''

        lr_info = self.check_LR_properties(self.adata, self.layer)
        counts_df, cell_thresholds = lr_info

        if not all(
            hasattr(self.adata.uns, attr) 
            for attr in ['received_ligands', 'received_ligands_tfl']
        ):
            self.adata = init_received_ligands(
                self.adata,
                radius=self.radius, 
                contact_distance=self.contact_distance, 
                cell_threshes=cell_thresholds
            )

        if len(self.lr['pairs']) > 0:
            
            self.adata.uns['ligand_receptor'] = self.ligands_receptors_interactions(
                self.adata.uns['received_ligands'][self.ligands], 
                get_filtered_df(counts_df, cell_thresholds, self.receptors)[self.receptors]
            )

        else:
            self.adata.uns['received_ligands'] = pd.DataFrame(index=self.adata.obs.index)
            self.adata.uns['ligand_receptor'] = pd.DataFrame(index=self.adata.obs.index)


        if len(self.tfl_pairs) > 0:
            self.adata.uns['ligand_regulator'] = self.ligand_regulators_interactions(
                self.adata.uns['received_ligands_tfl'][self.tfl_ligands], 
                self.adata.to_df(layer=self.layer)[self.tfl_regulators]
            )
        else:
            self.adata.uns['ligand_regulator'] = pd.DataFrame(index=self.adata.obs.index)

        self.train_df = self.adata.to_df(layer=self.layer)[
            [self.target_gene]+self.regulators] \
            .join(self.adata.uns['ligand_receptor']) \
            .join(self.adata.uns['ligand_regulator'])
        
        if len(self.lr_pairs) > 0:
            self.lr_pairs = self.lr_pairs[self.lr_pairs.isin(self.train_df.columns)]
        if len(self.tfl_pairs) > 0:
            self.tfl_pairs = [i for i in self.tfl_pairs if i in self.train_df.columns]
        
        self.ligands = []
        self.receptors = []
        self.tfl_regulators = []
        self.tfl_ligands = []
        
        for i in self.lr_pairs:
            lig, rec = i.split('$')
            self.ligands.append(lig)
            self.receptors.append(rec)
            
        for i in self.tfl_pairs:
            lig, reg = i.split('#')
            self.tfl_ligands.append(lig)
            self.tfl_regulators.append(reg)
            
        self.modulators = self.regulators + list(self.lr_pairs) + self.tfl_pairs
        self.modulators_genes = list(np.unique(
            self.regulators+self.ligands+self.receptors+self.tfl_regulators+self.tfl_ligands))

        assert len(self.ligands) == len(self.receptors)

        X = self.train_df.drop(columns=[self.target_gene]).values
        y = self.train_df[self.target_gene].values
        
        sp_maps = self.adata.obsm[self.sp_maps_key]

        assert sp_maps.shape[0] == X.shape[0] == y.shape[0] 
        return sp_maps, X, y 

    def fit(self, num_epochs=100, learning_rate=5e-3, batch_size=512, pbar=None):
        
        sp_maps, X, y = self.init_data()
        print(f'Using {self.sp_maps_key} as spatial maps')

        self.models = {}
        self.Xn = X
        self.yn = y
        self.sp_maps = sp_maps
        self.cell_indices = self.adata.obs.index.copy()

        if pbar is None:
            manager = enlighten.get_manager()
            pbar = manager.counter(
                total=sp_maps.shape[0]*num_epochs, 
                desc='Estimating Spatial Betas', unit='cells',
                color='green',
                auto_refresh=True
            )

        if num_epochs:
            print(f'Fitting {self.target_gene} with {len(self.modulators)} modulators')
            print(f'\t{len(self.regulators)} Transcription Factors')
            print(f'\t{len(self.lr_pairs)} Ligand-Receptor Pairs')
            print(f'\t{len(self.tfl_pairs)} TranscriptionFactor-Ligand Pairs')

        X_cell, y_cell = self.Xn, self.yn

        loader = DataLoader(
            PrefeaturizedTensorDataset(
                sp_maps,
                X_cell,
                y_cell
            ),
            batch_size=batch_size, shuffle=True
        )

        model = PrefeaturizedNicheNetwork(
                n_modulators = len(self.modulators), 
                spatial_dim=self.spatial_dim
            ).to(self.device)

        criterion = torch.nn.MSELoss()
        optimizer = torch.optim.Adam(model.parameters(), lr=learning_rate, weight_decay=0)
        
        # Early stopping parameters
        patience = 10
        best_loss = float('inf')
        patience_counter = 0

        for epoch in range(num_epochs):
            model.train()
            epoch_loss = 0
            all_y_true = []
            all_y_pred = []
            
            for batch in loader:
                spatial_maps, inputs, targets = [b.to(device) for b in batch]
                
                optimizer.zero_grad()
                outputs = model(spatial_maps, inputs)
                loss = criterion(outputs, targets)
                loss.backward()

                torch.nn.utils.clip_grad_norm_(model.parameters(), max_norm=2.0, norm_type=2)
                optimizer.step()
                
                epoch_loss += loss.item()
                all_y_true.extend(targets.cpu().detach().numpy())
                all_y_pred.extend(outputs.cpu().detach().numpy())

                pbar.desc = f'{self.target_gene}'
                pbar.update(len(targets))
    
            # Early stopping check
            avg_epoch_loss = epoch_loss / len(loader)

            if avg_epoch_loss < best_loss:
                best_loss = avg_epoch_loss
                patience_counter = 0
            else:
                patience_counter += 1
                
            if epoch > 50 and patience_counter >= patience:
                print(f'Early stopping triggered at epoch {epoch+1}')
                break

        if num_epochs:
            score = r2_score(all_y_true, all_y_pred)
            print(f'{score:.4f}')
        
        self.models = model
    
    @torch.no_grad()
    def get_betas(self):
        index_tracker = []
        betas = []
        indices = self.cell_indices
        index_tracker.extend(indices)
        sp_maps = torch.from_numpy(
            self.sp_maps).float().unsqueeze(1)
        b = self.models.get_betas(
            sp_maps.to(self.device),
        ).cpu().numpy()
        betas.extend(b)

        return pd.DataFrame(
            betas, 
            index=index_tracker, 
            columns=['beta0']+['beta_'+i for i in self.modulators]
        ).reindex(self.adata.obs.index)



class GeneProgramsEstimator(PrefeaturizedCellularProgramsEstimator):

    def __init__(self, adata, target_gene, mgs, layer='imputed_count', 
                 tf_ligand_cutoff=0.01, radius=100, contact_distance=30, scale_factor=1,
                 sp_maps_key='scGPT', use_ligands=False):

        if torch.backends.mps.is_available():
            device = torch.device("mps")
        elif torch.cuda.is_available():
            device = torch.device("cuda")
        else:
            device = torch.device("cpu")

        self.device = device
        self.adata = adata
        self.radius = radius
        self.contact_distance = contact_distance
        self.layer = layer
        self.mgs = mgs
        self.target_gene = target_gene
        self.tf_ligand_cutoff = tf_ligand_cutoff
        self.sp_maps_key = sp_maps_key
        self.spatial_dim = adata.obsm[sp_maps_key].shape[1]
        self.use_ligands = use_ligands

        ligand_mixtures = self.init_ligands_and_receptors(use_ligands=use_ligands)

        self.tfl_pairs = ligand_mixtures['tfl_pairs']        
        self.tfl_regulators = ligand_mixtures['tfl_regulators']
        self.tfl_ligands = ligand_mixtures['tfl_ligands']
        self.regulators = ligand_mixtures['regulators']

        # Treat all interactions as tfl pairs
        self.ligands = []
        self.receptors = []
        self.lr = {'pairs': []}
        self.lr_pairs = self.lr['pairs']
    

    def init_ligands_and_receptors(self, use_ligands=True):
    
        adata = self.adata
        species = 'mouse' if is_mouse_data(adata) else 'human'
        # df_ligrec = ct.pp.ligand_receptor_database(
        #     database='CellChat', 
        #     species=species, 
        #     signaling_type=None
        # ) 
        # df_ligrec.columns = ['ligand', 'receptor', 'pathway', 'signaling'] 
        
        df_ligrec = get_cellchat_db(species) 

        lr = expand_paired_interactions(df_ligrec)
        lr = lr[lr.ligand.isin(adata.var_names) &\
            (lr.receptor.isin(adata.var_names))]
        ligands = np.unique(lr.ligand)

        tfl_pairs = []
        tfl_regulators = []
        tfl_ligands = []

        grns = [genes for module, genes in self.mgs.items() if self.target_gene in genes]

        regulators = list(set(itertools.chain.from_iterable(grns)) - {self.target_gene})

        ligand_mixtures = edict()

        if use_ligands:

            for grn in grns:
                grn = [g for g in grn if g != self.target_gene]
                grn_pairs = pd.DataFrame(list(itertools.combinations(grn, 2)), columns=['gene1', 'gene2'])

                for g1, g2 in grn_pairs.values:
                    if g1 in ligands:
                        tfl_pairs.append(f'{g1}#{g2}')
                        tfl_regulators.append(g2)
                        tfl_ligands.append(g1)
                    if g2 in ligands:
                        tfl_pairs.append(f'{g2}#{g1}')
                        tfl_regulators.append(g1)
                        tfl_ligands.append(g2)

            ligand_mixtures['tfl_pairs'] = tfl_pairs
            ligand_mixtures['tfl_regulators'] = tfl_regulators
            ligand_mixtures['tfl_ligands'] = tfl_ligands

        else:

            ligand_mixtures['tfl_pairs'] = []
            ligand_mixtures['tfl_regulators'] = []
            ligand_mixtures['tfl_ligands'] = []

        ligand_mixtures['regulators'] = regulators
        return ligand_mixtures


import unittest
import numpy as np
import pandas as pd 
import os
# import sys
# sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'src')))
import tempfile
import shutil
import pickle
from unittest.mock import patch, MagicMock, Mock
import scanpy as sc
import anndata as ad
from scipy import sparse
from SpaceTravLR.spaceship import SpaceShip, Status, catch_and_retry, catch_errors

def quick_normalize(adata):
    adata.layers['raw_count'] = adata.X.copy()
    sc.pp.normalize_total(adata, target_sum=1e4)
    sc.pp.log1p(adata)
    adata.layers['normalized_count'] = adata.X.copy()
    return adata


def create_test_adata(n_cells=100, n_genes=50, species='human'):
    np.random.seed(42)
    X = np.random.rand(n_cells, n_genes)
    
    if species == 'mouse':
        gene_names = [f'Gene{i}' for i in range(n_genes)]
    else:
        gene_names = [f'GENE{i}' for i in range(n_genes)]
    
    adata = ad.AnnData(X=X)
    adata.var_names = gene_names
    adata.obs_names = [f'cell_{i}' for i in range(n_cells)]
    
    cell_types = np.random.choice(['TypeA', 'TypeB', 'TypeC'], size=n_cells)
    adata.obs['cell_type'] = cell_types
    
    spatial_coords = np.random.rand(n_cells, 2) * 1000
    adata.obsm['spatial'] = spatial_coords
    
    adata.layers['raw_count'] = X.copy()
    adata.layers['normalized_count'] = X.copy()
    
    return adata

def create_test_tfls(gene_names, n_ligands=10, n_tfs=10):
    cols = np.random.choice(gene_names, size=n_ligands, replace=False)
    tfs = np.random.choice(gene_names, size=n_tfs, replace=False)
    df_vals = np.random.rand(n_ligands, n_tfs)
    df = pd.DataFrame(df_vals, index=cols, columns=tfs)
    return df

class TestSpaceShip(unittest.TestCase):
    
    def setUp(self):
        self.temp_dir = tempfile.mkdtemp()
        self.original_cwd = os.getcwd()
        os.chdir(self.temp_dir)
        
    def tearDown(self):
        os.chdir(self.original_cwd)
        if os.path.exists(self.temp_dir):
            shutil.rmtree(self.temp_dir)
    
    def test_init(self):
        ship = SpaceShip()
        assert ship.name == 'AlienTissue'
        assert ship.status == Status.BORN
        del ship
        
        ship = SpaceShip(name='TestShip')
        assert ship.name == 'TestShip'
        assert ship.status == Status.BORN
        del ship
    
    def test_process_adata_basic(self):
        os.makedirs('output/input_data', exist_ok=True)
        
        adata = create_test_adata(species='human')
        ship = SpaceShip()
        
        with patch('SpaceTravLR.tools.utils.scale_adata') as mock_scale, \
             patch('SpaceTravLR.tools.utils.is_mouse_data', return_value=False), \
             patch('SpaceTravLR.tools.network.encode_labels') as mock_encode, \
             patch('SpaceTravLR.oracles.BaseTravLR.impute_clusterwise') as mock_impute:
            
            mock_encode.return_value = {'TypeA': 0, 'TypeB': 1, 'TypeC': 2}
            mock_scale.return_value = adata
            
            ship.process_adata_(adata, annot='cell_type')
            
            assert ship.annot == 'cell_type'
            assert ship.species == 'human'
            assert os.path.exists('output/input_data/_adata.h5ad')
            assert 'cell_type_int' in ship.adata.obs.columns
    
    def test_process_adata_missing_spatial(self):
        adata = create_test_adata()
        del adata.obsm['spatial']
        ship = SpaceShip()
        
        with self.assertRaises(AssertionError):
            ship.process_adata_(adata)
    
    def test_process_adata_missing_annot(self):
        adata = create_test_adata()
        ship = SpaceShip()
        
        with self.assertRaises(AssertionError):
            ship.process_adata_(adata, annot='nonexistent')
    
    
    def test_load_base_cell_thresholds(self):
        adata = create_test_adata()
        ship = SpaceShip()
        ship.adata = adata
        ship.species = 'human'
        
        mock_df = pd.DataFrame({
            'ligand': ['Lig1', 'Lig2'],
            'receptor': ['Rec1', 'Rec2']
        })
        
        mock_expanded = pd.DataFrame({
            'ligand': ['Lig1', 'Lig2'],
            'receptor': ['Rec1', 'Rec2']
        })
        
        with patch('SpaceTravLR.spaceship.get_cellchat_db', return_value=mock_df), \
             patch('SpaceTravLR.spaceship.expand_paired_interactions', return_value=mock_expanded):
            
            result = ship.load_base_cell_thresholds()
            
            assert isinstance(result, pd.DataFrame)
            assert len(result) == len(adata.obs_names)
            assert set(result.columns) == {'Lig1', 'Lig2', 'Rec1', 'Rec2'}
    
    def test_load_base_GRN_human(self):
        mock_df = pd.DataFrame({
            'gene_short_name': ['Gene1', 'Gene2', 'Gene3'],
            'TF1': [1, 0, 1],
            'TF2': [0, 1, 0],
            'peak_id': ['peak1', 'peak2', 'peak3']
        })
        
        with patch('pandas.read_parquet', return_value=mock_df):
            result = SpaceShip.load_base_GRN('human')

            print(result.head())
            
            assert isinstance(result, pd.DataFrame)
            assert 'source' in result.columns
            assert 'target' in result.columns
            assert 'coef_mean' in result.columns
            assert 'coef_abs' in result.columns
            assert 'p' in result.columns
            assert '-logp' in result.columns
    
    def test_load_base_GRN_mouse(self):
        mock_df = pd.DataFrame({
            'gene_short_name': ['Gene1', 'Gene2'],
            'TF1': [1, 0],
            'peak_id': ['peak1', 'peak2']
        })
        
        with patch('pandas.read_parquet', return_value=mock_df):
            result = SpaceShip.load_base_GRN('mouse')
            
            assert isinstance(result, pd.DataFrame)
    
    def test_load_base_GRN_invalid_species(self):
        with self.assertRaises(AssertionError):
            SpaceShip.load_base_GRN('invalid')
    
    def test_run_celloracle_(self):
        os.makedirs('output/input_data', exist_ok=True)
        
        adata = create_test_adata()
        adata.layers['raw_count'] = adata.X.copy()
        adata.obsm['X_umap'] = np.random.rand(len(adata), 2)
        
        ship = SpaceShip()
        ship.adata = adata
        ship.annot = 'cell_type'
        ship.species = 'human'
        
        mock_oracle = MagicMock()
        mock_links = MagicMock()
        mock_links.links_dict = {'TypeA': pd.DataFrame(), 'TypeB': pd.DataFrame()}
        mock_oracle.get_links.return_value = mock_links
        
        mock_co = MagicMock()
        mock_co.Oracle.return_value = mock_oracle
        mock_co.data.load_human_promoter_base_GRN.return_value = pd.DataFrame()
        
        with patch('SpaceTravLR.spaceship.sys.path'), \
             patch.dict('sys.modules', {'celloracle_tmp': mock_co}):
            ship.run_celloracle_(alpha=5)
            
            self.assertTrue(hasattr(ship, 'links'))
            self.assertTrue(os.path.exists('output/input_data/celloracle_links.pkl'))
    
    def test_run_celloracle_fallback(self):
        adata = create_test_adata()
        adata.layers['raw_count'] = adata.X.copy()
        adata.obsm['X_umap'] = np.random.rand(len(adata), 2)
        
        ship = SpaceShip()
        ship.adata = adata
        ship.annot = 'cell_type'
        ship.species = 'mouse'
        
        mock_oracle = MagicMock()
        mock_links = MagicMock()
        mock_links.links_dict = {'TypeA': pd.DataFrame()}
        mock_oracle.get_links.return_value = mock_links
        
        mock_co = MagicMock()
        mock_co.Oracle.return_value = mock_oracle
        mock_co.data.load_mouse_promoter_base_GRN.return_value = pd.DataFrame()
        
        with patch('SpaceTravLR.spaceship.sys.path'), \
             patch.dict('sys.modules', {'celloracle': None}), \
             patch('importlib.import_module', return_value=mock_co):
            try:
                ship.run_celloracle_()
            except (ImportError, AttributeError):
                pass

    def test_setup_new(self):
        adata = create_test_adata()
        ship = SpaceShip()
        
        with patch.object(ship, 'process_adata_') as mock_process, \
             patch.object(ship, 'run_celloracle_') as mock_celloracle, \
             patch.object(ship, 'run_commot_') as mock_commot, \
             patch.object(ship, 'get_nichenet_links_') as mock_nichenet:
            
            result = ship.setup_(adata, overwrite=False)
            
            assert ship.status == Status.BORED
            assert result == ship
            mock_process.assert_called_once()
            mock_celloracle.assert_called_once()
            mock_commot.assert_called_once()
            mock_nichenet.assert_called_once()
    
    def test_setup_existing_no_overwrite(self):
        adata = create_test_adata()
        ship = SpaceShip()
        
        os.makedirs('output', exist_ok=True)
        
        with patch.object(ship, 'process_adata_'):
            result = ship.setup_(adata, overwrite=False)
            
            assert ship.status == Status.FUBAR
            assert result is None
    
    def test_setup_existing_overwrite(self):
        adata = create_test_adata()
        ship = SpaceShip()
        
        os.makedirs('output', exist_ok=True)
        
        with patch.object(ship, 'process_adata_') as mock_process, \
             patch.object(ship, 'run_celloracle_') as mock_celloracle, \
             patch.object(ship, 'run_commot_') as mock_commot, \
             patch.object(ship, 'get_nichenet_links_') as mock_nichenet:
            
            result = ship.setup_(adata, overwrite=True)
            
            assert ship.status == Status.BORED
            mock_process.assert_called_once()
            mock_nichenet.assert_called_once()
    
    def test_spawn_worker(self):
        os.makedirs('output/logs', exist_ok=True)
        
        ship = SpaceShip(name='TestShip')
        
        mock_slurm = MagicMock()
        
        with patch('SpaceTravLR.spaceship.Slurm', return_value=mock_slurm), \
             patch('time.strftime', return_value='20240101_120000'):
            
            ship.spawn_worker(
                partition='test',
                clusters='cpu',
                gres='gpu:1',
                job_name='TestJob',
                lifespan=1,
                python_path='python3'
            )
            
            mock_slurm.sbatch.assert_called_once_with('python3 launch.py')
    
    def test_run_spacetravlr(self):
        os.makedirs('output/input_data', exist_ok=True)
        os.makedirs('output/betadata', exist_ok=True)
        
        ship = SpaceShip()
        
        adata = create_test_adata()
        adata.write_h5ad('output/input_data/_adata.h5ad')
        
        mock_links = {'TypeA': pd.DataFrame(), 'TypeB': pd.DataFrame()}
        with open('output/input_data/celloracle_links.pkl', 'wb') as f:
            pickle.dump(mock_links, f)

        mock_nichenet_links = create_test_tfls(adata.var_names, n_ligands=3, n_tfs=10)
        mock_nichenet_links.to_parquet('output/input_data/tflinks.parquet')
        
        mock_tflinks = pd.DataFrame({'ligand': ['Lig1'], 'target': ['Gene1'], 'weight': [0.5]})
        
        mock_space_travlr = MagicMock()
        mock_regulatory_factory = MagicMock()
        
        with patch('SpaceTravLR.oracles.SpaceTravLR', return_value=mock_space_travlr), \
             patch('SpaceTravLR.tools.network.RegulatoryFactory', return_value=mock_regulatory_factory):            
            ship.run_spacetravlr(
                max_epochs=10,
                learning_rate=1e-3,
                spatial_dim=32,
                batch_size=256,
                radius=200,
                contact_distance=50
            )
            
            mock_space_travlr.run.assert_called_once()
    
    def test_fit_alias(self):
        ship = SpaceShip()
        
        with patch.object(ship, 'run_spacetravlr') as mock_run:
            ship.fit(max_epochs=10)
            mock_run.assert_called_once_with(max_epochs=10)
    
    def test_is_everything_ok_success(self):
        ship = SpaceShip()
        
        os.makedirs('output/input_data', exist_ok=True)
        os.makedirs('output/betadata', exist_ok=True)
        os.makedirs('output/logs', exist_ok=True)
        
        adata = create_test_adata()
        adata.layers['normalized_count'] = adata.X.copy()
        adata.layers['imputed_count'] = adata.X.copy()
        adata.obsm['X_umap'] = np.random.rand(len(adata), 2)
        adata.obs['cell_type_int'] = np.random.randint(0, 3, len(adata))
        adata.write_h5ad('output/input_data/_adata.h5ad')
        
        with open('output/input_data/celloracle_links.pkl', 'wb') as f:
            pickle.dump({}, f)
        
        with open('launch.py', 'w') as f:
            f.write('# launch script\n')
        
        result = ship.is_everything_ok()
        assert result is True
    
    def test_is_everything_ok_missing_adata(self):
        ship = SpaceShip()
        
        os.makedirs('output/input_data', exist_ok=True)
        
        with self.assertRaisesRegex(AssertionError, "AnnData file not found"):
            ship.is_everything_ok()
    
    def test_is_everything_ok_missing_layers(self):
        ship = SpaceShip()
        
        os.makedirs('output/input_data', exist_ok=True)
        
        adata = create_test_adata()
        adata.write_h5ad('output/input_data/_adata.h5ad')
        
        with open('output/input_data/celloracle_links.pkl', 'wb') as f:
            pickle.dump({}, f)
        
        with self.assertRaisesRegex(AssertionError, "Imputed count layer not found"):
            ship.is_everything_ok()
    
    def test_catch_and_retry_decorator(self):
        call_count = [0]
        
        @catch_and_retry(retry=1)
        def failing_function():
            call_count[0] += 1
            raise ValueError("Test error")
        
        with self.assertRaises(ValueError):
            failing_function()
        
        self.assertEqual(call_count[0], 1)
    
    def test_catch_errors_alias(self):
        self.assertTrue(callable(catch_errors))
        
        call_count = [0]
        
        @catch_errors
        def test_function():
            call_count[0] += 1
            return "success"
        
        result = test_function()
        self.assertEqual(result, "success")
        self.assertEqual(call_count[0], 1)


if __name__ == '__main__':
    unittest.main()




import numpy as np
import torch 
import torch.nn as nn
import torch.nn.functional as F

device = torch.device(
    "mps" if torch.backends.mps.is_available() 
    else "cuda" if torch.cuda.is_available() 
    else "cpu"
)

def cosine_annealing(epoch, total_epochs):
    return (0.5 * (1 + np.cos(np.pi * (epoch - 0) / total_epochs))) 

class ViT(nn.Module):
    def __init__(self, betas, in_channels, spatial_dim, n_patches=4, n_blocks=4, hidden_d=16, n_heads=8):
        super().__init__()
        
        self.__version__ = 5.0
        
        self.betas = betas
        self.dim = betas.shape[0] # number of TFs
        self.in_channels = in_channels
        self.spatial_dim = spatial_dim
        # Attributes
        chw = (in_channels, spatial_dim, spatial_dim) # ( C , H , W )
        self.n_patches = n_patches
        self.n_blocks = n_blocks
        self.n_heads = n_heads
        self.hidden_d = hidden_d
        
        # Input and patches sizes
        assert chw[1] % n_patches == 0, "Input shape not entirely divisible by number of patches"
        assert chw[2] % n_patches == 0, "Input shape not entirely divisible by number of patches"
        self.patch_size = (chw[1] / n_patches, chw[2] / n_patches)

        self.input_d = int(chw[0] * self.patch_size[0] * self.patch_size[1])
        self.linear_mapper = nn.Linear(self.input_d, self.hidden_d)
        
        self.class_token = nn.Parameter(torch.rand(1, self.hidden_d))       # Consider removing
        self.pos_embed = nn.Parameter(get_positional_embeddings(self.n_patches ** 2 + 1, self.hidden_d))
        self.pos_embed.requires_grad = False

        self.cluster_embed = nn.Embedding(self.in_channels, self.hidden_d)
        
        self.blocks = nn.ModuleList(
            [ViTBlock(hidden_d, n_heads) for _ in range(n_blocks)])

        self.mlp = nn.Sequential(
            nn.Linear(self.hidden_d, 32),
            nn.GELU(),
            nn.Dropout(0.1),
            nn.Linear(32, 16),
            nn.GELU(),
            nn.Dropout(0.1),
            nn.Linear(16, self.dim)
        )

        self.alpha = nn.Parameter(torch.tensor(0, dtype=torch.float32), requires_grad=True)
        # self.alpha = 5e-2

        self.e = 0

    def forward(self, images, inputs_labels):
        n, c, h, w = images.shape 
        patches = patchify(images, self.n_patches).to(self.pos_embed.device)
        
        tokens = self.linear_mapper(patches) 
        tokens = torch.cat((self.class_token.expand(n, 1, -1), tokens), dim=1)
        out = tokens + self.pos_embed.repeat(n, 1, 1)
        
        for j, block in enumerate(self.blocks):
            out = block(out)
            
        out = out[:, 0]

        emb = self.cluster_embed(inputs_labels) * 1
        # out = torch.concat([out, emb], dim=1)
        out = out + emb
        betas = self.mlp(out)

        return betas
        
    
    def get_att_weights(self, images):
        n, c, h, w = images.shape 
        patches = patchify(images, self.n_patches).to(self.pos_embed.device)
        
        tokens = self.linear_mapper(patches)
        tokens = torch.cat((self.class_token.expand(n, 1, -1), tokens), dim=1)
        out = tokens + self.pos_embed.repeat(n, 1, 1)
        
        att_weights = []   # (n_blocks, batch, n_heads, seqs, seqs) where seqs is flattened patches
        for block in self.blocks:
            out, att = block.forward_att(out)
            att_weights.append(att)
        
        return att_weights




    def __str__(self):
        # return ''
        return f'VisionTransformer(in_channels={self.in_channels}, spatial_dim={self.spatial_dim}, n_patches={self.n_patches}, n_blocks={self.n_blocks}, hidden_d={self.hidden_d}, n_heads={self.n_heads})'

    def __repr__(self):
        return self.__str__()
        
        



class SkeletonViT(nn.Module):
    def __init__(self, n_modulators, in_channels, anchors=None, spatial_dim=64, n_patches=4, n_blocks=4, hidden_d=16, n_heads=8):
        super().__init__()
        
        self.dim = n_modulators+1
        if anchors is None:
            anchors = np.ones(self.dim)

        self.anchors = torch.from_numpy(anchors).float().to(device)
        
        self.in_channels = in_channels
        self.spatial_dim = spatial_dim
        
        chw = (in_channels, spatial_dim, spatial_dim) # ( C , H , W )
        self.n_patches = n_patches
        self.n_blocks = n_blocks
        self.n_heads = n_heads
        self.hidden_d = hidden_d
        
        # Input and patches sizes
        assert chw[1] % n_patches == 0, "Input shape not entirely divisible by number of patches"
        assert chw[2] % n_patches == 0, "Input shape not entirely divisible by number of patches"
        self.patch_size = (chw[1] / n_patches, chw[2] / n_patches)

        self.input_d = int(chw[0] * self.patch_size[0] * self.patch_size[1])
        self.linear_mapper = nn.Linear(self.input_d, self.hidden_d)
        
        self.class_token = nn.Parameter(torch.rand(1, self.hidden_d))       # Consider removing
        self.pos_embed = nn.Parameter(get_positional_embeddings(self.n_patches ** 2 + 1, self.hidden_d))
        self.pos_embed.requires_grad = False

        self.cluster_embed = nn.Embedding(self.in_channels, self.hidden_d)
        
        self.blocks = nn.ModuleList(
            [ViTBlock(hidden_d, n_heads) for _ in range(n_blocks)])

        self.mlp = nn.Sequential(
            nn.Linear(self.hidden_d, 32),
            nn.GELU(),
            nn.Dropout(0.1),
            nn.Linear(32, 16),
            nn.GELU(),
            nn.Dropout(0.1),
            nn.Linear(16, self.dim)
        )

        self.alpha = nn.Parameter(torch.tensor(0, dtype=torch.float32), requires_grad=True)
        # self.alpha = 5e-2

        self.e = 0

    def forward(self, images, inputs_labels):
        n, c, h, w = images.shape 
        patches = patchify(images, self.n_patches).to(self.pos_embed.device)
        
        tokens = self.linear_mapper(patches) 
        tokens = torch.cat((self.class_token.expand(n, 1, -1), tokens), dim=1)
        out = tokens + self.pos_embed.repeat(n, 1, 1)
        
        for j, block in enumerate(self.blocks):
            out = block(out)
            
        out = out[:, 0]

        emb = self.cluster_embed(inputs_labels) * 1
        # out = torch.concat([out, emb], dim=1)
        out = out + emb
        betas = self.mlp(out)

        return betas
        
    
    def get_att_weights(self, images):
        n, c, h, w = images.shape 
        patches = patchify(images, self.n_patches).to(self.pos_embed.device)
        
        tokens = self.linear_mapper(patches)
        tokens = torch.cat((self.class_token.expand(n, 1, -1), tokens), dim=1)
        out = tokens + self.pos_embed.repeat(n, 1, 1)
        
        att_weights = []   # (n_blocks, batch, n_heads, seqs, seqs) where seqs is flattened patches
        for block in self.blocks:
            out, att = block.forward_att(out)
            att_weights.append(att)
        
        return att_weights


    def __repr__(self):
        return self.__str__()
        
        


    



    

class ViTBlock(nn.Module):
    def __init__(self, hidden_d, n_heads, mlp_ratio=4):
        super(ViTBlock, self).__init__()
        self.hidden_d = hidden_d
        self.n_heads = n_heads

        self.norm1 = nn.LayerNorm(hidden_d)
        self.mhsa = MSA(hidden_d, n_heads)
        self.norm2 = nn.LayerNorm(hidden_d)
        self.mlp = nn.Sequential(
            nn.Linear(hidden_d, mlp_ratio * hidden_d),
            nn.GELU(),
            nn.Linear(mlp_ratio * hidden_d, hidden_d)
        )

    def forward(self, x):
        out = x + self.mhsa(self.norm1(x))
        out = out + self.mlp(self.norm2(out))
        return out
    
    def forward_att(self, x):
        out, att = self.mhsa.forward_att(self.norm1(x))
        out += x
        out = out + self.mlp(self.norm2(out))
        return out, att


class MSA(nn.Module):
    def __init__(self, d, n_heads=2):
        super().__init__()
        self.d = d
        self.n_heads = n_heads

        assert d % n_heads == 0, f"Can't divide dimension {d} into {n_heads} heads"

        d_head = int(d / n_heads)
        self.q_mappings = nn.ModuleList([nn.Linear(d_head, d_head) for _ in range(self.n_heads)])
        self.k_mappings = nn.ModuleList([nn.Linear(d_head, d_head) for _ in range(self.n_heads)])
        self.v_mappings = nn.ModuleList([nn.Linear(d_head, d_head) for _ in range(self.n_heads)])
        self.d_head = d_head
        self.softmax = nn.Softmax(dim=-1)

    # def forward(self, sequences):
    #     # Sequences has shape (N, seq_length, token_dim)
    #     # We go into shape    (N, seq_length, n_heads, token_dim / n_heads)
    #     # And come back to    (N, seq_length, item_dim)  (through concatenation)
    #     result = []
    #     for sequence in sequences:
    #         seq_result = []
    #         for head in range(self.n_heads):
    #             q_mapping = self.q_mappings[head]
    #             k_mapping = self.k_mappings[head]
    #             v_mapping = self.v_mappings[head]

    #             seq = sequence[:, head * self.d_head: (head + 1) * self.d_head]
    #             q, k, v = q_mapping(seq), k_mapping(seq), v_mapping(seq)

    #             attention = self.softmax(q @ k.T / (self.d_head ** 0.5))
    #             seq_result.append(attention @ v)
    #         result.append(torch.hstack(seq_result))
    #     return torch.cat([torch.unsqueeze(r, dim=0) for r in result])
    


    def forward(self, sequences):
        N, seq_length, token_dim = sequences.shape
        sequences = sequences.view(N, seq_length, self.n_heads, self.d_head)
        q = torch.stack([q_map(sequences[:,:,i,:]) for i, q_map in enumerate(self.q_mappings)], dim=2)
        k = torch.stack([k_map(sequences[:,:,i,:]) for i, k_map in enumerate(self.k_mappings)], dim=2)
        v = torch.stack([v_map(sequences[:,:,i,:]) for i, v_map in enumerate(self.v_mappings)], dim=2)

        q = q.transpose(1, 2)
        k = k.transpose(1, 2)
        v = v.transpose(1, 2)
        
        output = F.scaled_dot_product_attention(q, k, v)
        return output.transpose(1, 2).contiguous().view(N, seq_length, -1)
    
    # def forward_att(self, sequences):
    #     result = []
    #     atts = []

    #     for sequence in sequences:
    #         seq_result = []
    #         att_result = []

    #         for head in range(self.n_heads):
    #             q_mapping = self.q_mappings[head]
    #             k_mapping = self.k_mappings[head]
    #             v_mapping = self.v_mappings[head]

    #             seq = sequence[:, head * self.d_head: (head + 1) * self.d_head]
    #             q, k, v = q_mapping(seq), k_mapping(seq), v_mapping(seq)

    #             attention = self.softmax(q @ k.T / (self.d_head ** 0.5))
    #             seq_result.append(attention @ v)
    #             att_result.append(attention) 
            
    #         atts.append(torch.stack(att_result, dim=0))
    #         result.append(torch.hstack(seq_result))

    #     outs = torch.cat([torch.unsqueeze(r, dim=0) for r in result])

    #     return outs, atts

    def forward_att(self, sequences):
        N, seq_length, token_dim = sequences.shape
        sequences = sequences.view(N, seq_length, self.n_heads, self.d_head)
        
        q = torch.stack([q_map(sequences[:,:,i,:]) for i, q_map in enumerate(self.q_mappings)], dim=2)
        k = torch.stack([k_map(sequences[:,:,i,:]) for i, k_map in enumerate(self.k_mappings)], dim=2)
        v = torch.stack([v_map(sequences[:,:,i,:]) for i, v_map in enumerate(self.v_mappings)], dim=2)

        q = q.transpose(1, 2)
        k = k.transpose(1, 2)
        v = v.transpose(1, 2)
        
        attention = F.scaled_dot_product_attention(q, k, v, attn_mask=None, dropout_p=0.0, is_causal=False)
        
        outs = attention.transpose(1, 2).contiguous().view(N, seq_length, -1)
        atts = attention.softmax(dim=-1)  # Extract attention weights
        
        return outs, atts


def get_positional_embeddings(sequence_length, d):
    result = torch.ones(sequence_length, d)
    for i in range(sequence_length):
        for j in range(d):
            result[i][j] = np.sin(i / (10000 ** (j / d))) if j % 2 == 0 else np.cos(i / (10000 ** ((j - 1) / d)))
    return result

def patchify(images, n_patches):
    n, c, h, w = images.shape

    assert h == w, "Patchify method is implemented for square images only."

    patch_size = h // n_patches

    patches = images.unfold(2, patch_size, patch_size).unfold(3, patch_size, patch_size)
    patches = patches.permute(0, 2, 3, 1, 4, 5).reshape(n, n_patches**2, c * patch_size**2)

    return patches



import sys
import os
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'src')))

import numpy as np
import pandas as pd
import torch
import pytest
from unittest import TestCase
from unittest.mock import patch, MagicMock, Mock
from anndata import AnnData
from sklearn.metrics import r2_score
import pickle
import tempfile
import shutil
import matplotlib

from SpaceTravLR.models.parallel_estimators import SpatialCellularProgramsEstimator
from SpaceTravLR.models.pixel_attention import CellularNicheNetwork
from SpaceTravLR.tools.network import RegulatoryFactory


class MockRegulatoryFactory:
    def __init__(self, *args, **kwargs):
        pass
    
    def get_regulators(self, adata, target_gene):
        return ['TF1', 'TF2', 'TF3']

    def get_ligands(self, *args, **kwargs):
        return ['ligand1', 'ligand2']

    def get_receptors(self, *args, **kwargs):
        return ['receptor1', 'receptor2']

    def get_tfl_ligands(self, *args, **kwargs):
        return ['ligand1']

    def get_tfl_regulators(self, *args, **kwargs):
        return ['TF1']


class MockCellularNicheNetwork:
    def __init__(self, *args, **kwargs):
        self.device = torch.device('cpu')
        self.parameters = lambda: [torch.nn.Parameter(torch.randn(1))]
        self.anchors = torch.tensor([0.1, 0.2, 0.3, 0.4, 0.5, 0.6])
        
    def __call__(self, *args, **kwargs):
        # Return a tensor when called
        return torch.randn(10)
    
    def to(self, *args, **kwargs):
        return self
    
    def train(self):
        return self
    
    def eval(self):
        return self

    def get_betas(self):
        return torch.tensor([0.1, 0.2, 0.3, 0.4, 0.5])


class SpatialCellularProgramsEstimatorTest(TestCase):
    
    def setUp(self):
        # Set random seed for reproducibility
        np.random.seed(42)
        torch.manual_seed(42)
        
        # Force CPU usage for all operations
        self.device = torch.device('cpu')
        
        # Create synthetic AnnData object
        self.n_cells = 100
        self.n_genes = 50
        
        # Create synthetic gene expression data
        X = np.random.rand(self.n_cells, self.n_genes)
        
        # Create synthetic spatial coordinates
        spatial_coords = np.random.rand(self.n_cells, 2) * 1000
        
        # Create gene names including our target and regulatory genes
        gene_names = [f'gene_{i}' for i in range(self.n_genes)]
        gene_names[0] = 'target_gene'  # Our target gene
        gene_names[1:4] = ['TF1', 'TF2', 'TF3']  # TFs
        gene_names[4:8] = ['ligand1', 'ligand2', 'receptor1', 'receptor2']  # L-R pairs
        
        # Create cluster annotations
        clusters = np.random.randint(0, 3, size=self.n_cells)
        
        # Create AnnData object
        self.adata = AnnData(X=X)
        self.adata.var_names = gene_names
        self.adata.obs_names = [f'cell_{i}' for i in range(self.n_cells)]
        self.adata.obs['rctd_cluster'] = clusters
        self.adata.obs['cell_type_int'] = clusters
        
        self.adata.obsm['spatial'] = spatial_coords
        
        # Add imputed count layer
        self.adata.layers['imputed_count'] = X.copy()
        
        # Create temporary directory for model export/import testing
        self.temp_dir = tempfile.mkdtemp()
    
    def tearDown(self):
        # Clean up temporary directory
        shutil.rmtree(self.temp_dir)
    
    # @patch('SpaceTravLR.models.parallel_estimators.RegulatoryFactory', MockRegulatoryFactory)
    # def test_initialization(self):
    #     """Test that the estimator initializes correctly with valid parameters."""
    #     estimator = SpatialCellularProgramsEstimator(
    #         adata=self.adata,
    #         target_gene='target_gene',
    #         cluster_annot='rctd_cluster',
    #         layer='imputed_count',
    #         radius=200,
    #         contact_distance=50,
    #         tf_ligand_cutoff=0.01,
    #         colinks_path='dummy_path'
    #     )
        
    #     # Check that the estimator was initialized correctly
    #     self.assertEqual(estimator.target_gene, 'target_gene')
    #     self.assertEqual(estimator.cluster_annot, 'rctd_cluster')
    #     self.assertEqual(estimator.layer, 'imputed_count')
    #     self.assertEqual(estimator.radius, 200)
    #     self.assertEqual(estimator.contact_distance, 50)
    #     self.assertEqual(estimator.spatial_dim, 64)  # Default value
        
    #     # Check that regulators were set correctly
    #     self.assertIsNotNone(estimator.regulators)
    #     self.assertIsInstance(estimator.regulators, list)
        
    #     # Check that ligands and receptors were initialized
    #     self.assertIsNotNone(estimator.ligands)
    #     self.assertIsNotNone(estimator.receptors)
    
    @patch('SpaceTravLR.models.parallel_estimators.RegulatoryFactory', MockRegulatoryFactory)
    @patch('SpaceTravLR.tools.network.get_cellchat_db')
    @patch('SpaceTravLR.models.parallel_estimators.init_ligands_and_receptors')
    def test_init_ligands_and_receptors(self, mock_init_ligands, mock_get_cellchat_db):
        """Test the initialization of ligands and receptors."""
        # Mock the CellChat database
        mock_df_ligrec = pd.DataFrame({
            'ligand': ['ligand1', 'ligand2'],
            'receptor': ['receptor1', 'receptor2'],
            'pathway': ['pathway1', 'pathway2'],
            'signaling': ['Secreted Signaling', 'Cell-Cell Contact']
        })
        
        mock_get_cellchat_db.return_value = mock_df_ligrec
        
        # Create a mock return value for init_ligands_and_receptors
        from easydict import EasyDict as edict
        mock_ligand_mixtures = edict()
        mock_ligand_mixtures.lr = {'pairs': [('ligand1', 'receptor1'), ('ligand2', 'receptor2')]}
        mock_ligand_mixtures.ligands = ['ligand1', 'ligand2']
        mock_ligand_mixtures.receptors = ['receptor1', 'receptor2']
        mock_ligand_mixtures.tfl_pairs = [('TF1', 'ligand1')]
        mock_ligand_mixtures.tfl_regulators = ['TF1']
        mock_ligand_mixtures.tfl_ligands = ['ligand1']
        
        mock_init_ligands.return_value = mock_ligand_mixtures
        
        # Create the estimator
        estimator = SpatialCellularProgramsEstimator(
            adata=self.adata,
            target_gene='target_gene',
            cluster_annot='rctd_cluster',
            layer='imputed_count',
            radius=200,
            contact_distance=50,
            tf_ligand_cutoff=0.1,
            colinks_path='dummy_path'
        )
        
        # Test that the ligands and receptors are initialized correctly
        self.assertEqual(estimator.regulators, ['TF1', 'TF2', 'TF3'])  # This comes from MockRegulatoryFactory
        self.assertEqual(estimator.ligands, ['ligand1', 'ligand2'])
        self.assertEqual(estimator.receptors, ['receptor1', 'receptor2'])
        self.assertEqual(estimator.tfl_ligands, ['ligand1'])
        self.assertEqual(estimator.tfl_regulators, ['TF1'])
    
    @patch('SpaceTravLR.models.parallel_estimators.RegulatoryFactory', MockRegulatoryFactory)
    @patch('SpaceTravLR.tools.network.get_cellchat_db')
    @patch('SpaceTravLR.models.parallel_estimators.received_ligands')
    def test_init_data(self, mock_received_ligands, mock_get_cellchat_db):
        """Test the data initialization process."""
        # Mock the CellChat database
        mock_df_ligrec = pd.DataFrame({
            'ligand': ['ligand1', 'ligand2'],
            'receptor': ['receptor1', 'receptor2'],
            'pathway': ['pathway1', 'pathway2'],
            'signaling': ['Secreted Signaling', 'Cell-Cell Contact']
        })
        mock_get_cellchat_db.return_value = mock_df_ligrec
        
        # Mock the nichenet_lt dataframe
        mock_nichenet_lt = pd.DataFrame(
            np.random.rand(3, 2),
            index=['TF1', 'TF2', 'TF3'],
            columns=['ligand1', 'ligand2']
        )
        
        # Mock the received_ligands function
        mock_received_ligands.return_value = pd.DataFrame(
            np.random.rand(self.n_cells, 2),
            index=self.adata.obs.index,
            columns=['ligand1', 'ligand2']
        )
        
        # Patch the pd.read_parquet to return our mock dataframe
        with patch('pandas.read_parquet', return_value=mock_nichenet_lt):
            estimator = SpatialCellularProgramsEstimator(
                adata=self.adata,
                target_gene='target_gene',
                layer='imputed_count',
                radius=200,
                contact_distance=50,
                tf_ligand_cutoff=0.01,
                colinks_path='dummy_path'
            )
            
            # Initialize data
            sp_maps, X, y, cluster_labels = estimator.init_data()
        
        # Check that the data was initialized correctly
        self.assertIsNotNone(sp_maps)
        self.assertIsNotNone(X)
        self.assertIsNotNone(y)
        self.assertIsNotNone(cluster_labels)
        
        # Check shapes
        self.assertEqual(sp_maps.shape[0], self.n_cells)
        self.assertEqual(X.shape[0], self.n_cells)
        self.assertEqual(y.shape[0], self.n_cells)
        self.assertEqual(cluster_labels.shape[0], self.n_cells)
        
        # Check that the AnnData object was updated with the new data
        self.assertIn('received_ligands', self.adata.uns)
        self.assertIn('ligand_receptor', self.adata.uns)
        self.assertIn('spatial_maps', self.adata.obsm)
        self.assertIn('spatial_features', self.adata.obsm)
    
    def test_ligands_receptors_interactions(self):
        """Test the static method for computing ligand-receptor interactions."""
        # Create test dataframes
        n_samples = 10
        n_features = 5
        index = [f'cell_{i}' for i in range(n_samples)]
        columns_lig = [f'ligand_{i}' for i in range(n_features)]
        columns_rec = [f'receptor_{i}' for i in range(n_features)]
        
        received_ligands_df = pd.DataFrame(
            np.random.rand(n_samples, n_features),
            index=index,
            columns=columns_lig
        )
        
        receptor_gex_df = pd.DataFrame(
            np.random.rand(n_samples, n_features),
            index=index,
            columns=columns_rec
        )
        
        # Compute interactions
        lr_interactions = SpatialCellularProgramsEstimator.ligands_receptors_interactions(
            received_ligands_df, receptor_gex_df
        )
        
        # Check that the result is correct
        self.assertIsInstance(lr_interactions, pd.DataFrame)
        self.assertEqual(lr_interactions.shape, (n_samples, n_features))
        self.assertEqual(lr_interactions.index.tolist(), index)
        
        # Check that the column names are correctly formatted
        expected_columns = [f'{l}${r}' for l, r in zip(columns_lig, columns_rec)]
        self.assertEqual(lr_interactions.columns.tolist(), expected_columns)
        
        # Check that the values are correct (element-wise multiplication)
        expected_values = received_ligands_df.values * receptor_gex_df.values
        np.testing.assert_array_almost_equal(lr_interactions.values, expected_values)
    
    def test_ligand_regulators_interactions(self):
        """Test the static method for computing ligand-regulator interactions."""
        # Create test dataframes
        n_samples = 10
        n_features = 5
        index = [f'cell_{i}' for i in range(n_samples)]
        columns_lig = [f'ligand_{i}' for i in range(n_features)]
        columns_reg = [f'regulator_{i}' for i in range(n_features)]
        
        received_ligands_df = pd.DataFrame(
            np.random.rand(n_samples, n_features),
            index=index,
            columns=columns_lig
        )
        
        regulator_gex_df = pd.DataFrame(
            np.random.rand(n_samples, n_features),
            index=index,
            columns=columns_reg
        )
        
        # Compute interactions
        ltf_interactions = SpatialCellularProgramsEstimator.ligand_regulators_interactions(
            received_ligands_df, regulator_gex_df
        )
        
        # Check that the result is correct
        self.assertIsInstance(ltf_interactions, pd.DataFrame)
        self.assertEqual(ltf_interactions.shape, (n_samples, n_features))
        self.assertEqual(ltf_interactions.index.tolist(), index)
        
        # Check that the column names are correctly formatted
        expected_columns = [f'{l}#{r}' for l, r in zip(columns_lig, columns_reg)]
        self.assertEqual(ltf_interactions.columns.tolist(), expected_columns)
        
        # Check that the values are correct (element-wise multiplication)
        expected_values = received_ligands_df.values * regulator_gex_df.values
        np.testing.assert_array_almost_equal(ltf_interactions.values, expected_values)
    
    @patch('SpaceTravLR.models.parallel_estimators.RegulatoryFactory', MockRegulatoryFactory)
    @patch('SpaceTravLR.tools.network.get_cellchat_db')
    @patch('SpaceTravLR.models.parallel_estimators.received_ligands')
    @patch('SpaceTravLR.models.parallel_estimators.CellularNicheNetwork', MockCellularNicheNetwork)
    @patch('SpaceTravLR.models.parallel_estimators.torch.optim.Adam')
    @patch('SpaceTravLR.models.parallel_estimators.DataLoader')
    @patch('SpaceTravLR.models.parallel_estimators.enlighten.get_manager')
    @patch('SpaceTravLR.models.parallel_estimators.RotatedTensorDataset')
    def test_fit_and_get_betas(self, mock_dataset, mock_manager, mock_dataloader, mock_adam, mock_received_ligands, mock_get_cellchat_db):
        """Test the fit method and getting betas."""
        # Mock the CellChat database
        mock_df_ligrec = pd.DataFrame({
            'ligand': ['ligand1', 'ligand2'],
            'receptor': ['receptor1', 'receptor2'],
            'pathway': ['pathway1', 'pathway2'],
            'signaling': ['Secreted Signaling', 'Cell-Cell Contact']
        })
        mock_get_cellchat_db.return_value = mock_df_ligrec
        
        # Mock the nichenet_lt dataframe
        mock_nichenet_lt = pd.DataFrame(
            np.random.rand(3, 2),
            index=['TF1', 'TF2', 'TF3'],
            columns=['ligand1', 'ligand2']
        )
        
        # Mock the received_ligands function
        mock_received_ligands.return_value = pd.DataFrame(
            np.random.rand(self.n_cells, 2),
            index=self.adata.obs.index,
            columns=['ligand1', 'ligand2']
        )
        
        # Mock the DataLoader
        mock_dataloader_instance = MagicMock()
        mock_dataloader.return_value = mock_dataloader_instance
        mock_dataloader_instance.__iter__.return_value = [
            (torch.randn(10, 1, 64, 64), torch.randn(10, 5), torch.randn(10), torch.randn(10, 3))
        ]
        
        # Mock the enlighten manager
        mock_counter = MagicMock()
        mock_manager.return_value.counter.return_value = mock_counter
        
        # Mock the optimizer
        mock_optimizer = MagicMock()
        mock_adam.return_value = mock_optimizer
        
        # Create a custom estimator class for testing
        class TestEstimator(SpatialCellularProgramsEstimator):
            def fit(self, num_epochs=10, threshold_lambda=1e-4, learning_rate=2e-4, batch_size=512,
                    use_ARD=False, pbar=None, discard=50, use_bayesian=True, score_threshold=0.1):
                # Set up the model
                self.regulators = ['TF1', 'TF2', 'TF3']
                self.ligands = ['ligand1', 'ligand2']
                self.receptors = ['receptor1', 'receptor2']
                self.tfl_ligands = ['ligand1']
                self.tfl_regulators = ['TF1']
                self.lr_pairs = [('ligand1', 'receptor1'), ('ligand2', 'receptor2')]
                self.tfl_pairs = [('TF1', 'ligand1')]
                self.modulators = self.regulators + self.lr_pairs + self.tfl_pairs
                
                self.model = MockCellularNicheNetwork()
                self.models = {'cluster1': self.model}
                return self
            
            def get_betas(self):
                return pd.DataFrame({
                    'beta': [0.1, 0.2, 0.3, 0.4, 0.5],
                    'modulator': ['TF1', 'TF2', 'TF3', ('ligand1', 'receptor1'), ('ligand2', 'receptor2')]
                })

        # Patch the pd.read_parquet to return our mock dataframe
        with patch('pandas.read_parquet', return_value=mock_nichenet_lt):
            # Create the test estimator
            estimator = TestEstimator(
                adata=self.adata,
                target_gene='target_gene',
                cluster_annot='rctd_cluster',
                layer='imputed_count',
                radius=200,
                contact_distance=50,
                tf_ligand_cutoff=0.01,
                colinks_path='dummy_path'
            )
            
            # Call fit
            estimator.fit(num_epochs=1, batch_size=32, use_bayesian=True)
            
            # Test getting betas
            betas = estimator.get_betas()
            
            # Check that betas is a pandas DataFrame
            self.assertIsInstance(betas, pd.DataFrame)
            
            # Check that the betas dataframe has the expected structure
            self.assertEqual(betas.shape[0], 5)  # 5 modulators (3 TFs + 2 LR pairs)
    
    @patch('SpaceTravLR.models.parallel_estimators.RegulatoryFactory', MockRegulatoryFactory)
    @patch('SpaceTravLR.tools.network.get_cellchat_db')
    @patch('SpaceTravLR.models.parallel_estimators.received_ligands')
    @patch('SpaceTravLR.models.parallel_estimators.CellularNicheNetwork', MockCellularNicheNetwork)
    @patch('SpaceTravLR.models.parallel_estimators.torch.optim.Adam')
    @patch('SpaceTravLR.models.parallel_estimators.DataLoader')
    @patch('SpaceTravLR.models.parallel_estimators.enlighten.get_manager')
    @patch('SpaceTravLR.models.parallel_estimators.RotatedTensorDataset')
    def test_export_and_load(self, mock_dataset, mock_manager, mock_dataloader, mock_adam, mock_received_ligands, mock_get_cellchat_db):
        """Test exporting and loading the model."""
        # Mock the CellChat database
        mock_df_ligrec = pd.DataFrame({
            'ligand': ['ligand1', 'ligand2'],
            'receptor': ['receptor1', 'receptor2'],
            'pathway': ['pathway1', 'pathway2'],
            'signaling': ['Secreted Signaling', 'Cell-Cell Contact']
        })
        mock_get_cellchat_db.return_value = mock_df_ligrec
        
        # Mock the nichenet_lt dataframe
        mock_nichenet_lt = pd.DataFrame(
            np.random.rand(3, 2),
            index=['TF1', 'TF2', 'TF3'],
            columns=['ligand1', 'ligand2']
        )
        
        # Mock the received_ligands function
        mock_received_ligands.return_value = pd.DataFrame(
            np.random.rand(self.n_cells, 2),
            index=self.adata.obs.index,
            columns=['ligand1', 'ligand2']
        )
        
        # Mock the DataLoader
        mock_dataloader_instance = MagicMock()
        mock_dataloader.return_value = mock_dataloader_instance
        mock_dataloader_instance.__iter__.return_value = [
            (torch.randn(10, 1, 64, 64), torch.randn(10, 5), torch.randn(10), torch.randn(10, 3))
        ]
        
        # Mock the enlighten manager
        mock_counter = MagicMock()
        mock_manager.return_value.counter.return_value = mock_counter
        
        # Mock the optimizer
        mock_optimizer = MagicMock()
        mock_adam.return_value = mock_optimizer
        
        # Create a custom estimator class for testing
        class TestEstimator(SpatialCellularProgramsEstimator):
            def fit(self, num_epochs=10, threshold_lambda=1e-4, learning_rate=2e-4, batch_size=512,
                    use_ARD=False, pbar=None, discard=50, use_bayesian=True, score_threshold=0.1):
                # Set up the model
                self.regulators = ['TF1', 'TF2', 'TF3']
                self.ligands = ['ligand1', 'ligand2']
                self.receptors = ['receptor1', 'receptor2']
                self.tfl_ligands = ['ligand1']
                self.tfl_regulators = ['TF1']
                self.lr_pairs = [('ligand1', 'receptor1'), ('ligand2', 'receptor2')]
                self.tfl_pairs = [('TF1', 'ligand1')]
                self.modulators = self.regulators + self.lr_pairs + self.tfl_pairs
                
                self.model = MockCellularNicheNetwork()
                self.models = {'cluster1': self.model}
                return self
            
            def export(self, path):
                # Mock implementation
                return

        # Create a custom SpatialCellularProgramsEstimator class for testing load
        class TestLoadEstimator(SpatialCellularProgramsEstimator):
            def load(self, path):
                # Mock implementation
                self.regulators = ['TF1', 'TF2', 'TF3']
                self.ligands = ['ligand1', 'ligand2']
                self.receptors = ['receptor1', 'receptor2']
                self.tfl_ligands = ['ligand1']
                self.tfl_regulators = ['TF1']
                return self

        # Patch the pd.read_parquet to return our mock dataframe
        with patch('pandas.read_parquet', return_value=mock_nichenet_lt):
            # Create the test estimator
            estimator = TestEstimator(
                adata=self.adata,
                target_gene='target_gene',
                cluster_annot='rctd_cluster',
                layer='imputed_count',
                radius=200,
                contact_distance=50,
                tf_ligand_cutoff=0.01,
                colinks_path='dummy_path'
            )
            
            # Call fit
            estimator.fit(num_epochs=1, batch_size=32, use_bayesian=True)
            
            # Create a temporary directory for export/load testing
            with tempfile.TemporaryDirectory() as temp_dir:
                model_path = os.path.join(temp_dir, 'model.pt')
                
                # Export the model
                estimator.export(model_path)
                
                # Create a new estimator
                new_estimator = TestLoadEstimator(
                    adata=self.adata,
                    target_gene='target_gene',
                    cluster_annot='rctd_cluster',
                    layer='imputed_count',
                    radius=200,
                    contact_distance=50,
                    tf_ligand_cutoff=0.01,
                    colinks_path='dummy_path'
                )
                
                # Load the model
                new_estimator.load(model_path)
                
                # Check that the attributes were loaded correctly
                self.assertEqual(new_estimator.regulators, ['TF1', 'TF2', 'TF3'])
                self.assertEqual(new_estimator.ligands, ['ligand1', 'ligand2'])
                self.assertEqual(new_estimator.receptors, ['receptor1', 'receptor2'])
                self.assertEqual(new_estimator.tfl_ligands, ['ligand1'])
                self.assertEqual(new_estimator.tfl_regulators, ['TF1'])
    
    @patch('SpaceTravLR.models.parallel_estimators.RegulatoryFactory', MockRegulatoryFactory)
    @patch('SpaceTravLR.tools.network.get_cellchat_db')
    @patch('SpaceTravLR.models.parallel_estimators.WordCloud')
    @patch('matplotlib.pyplot')
    def test_plot_modulators(self, mock_plt, mock_wordcloud, mock_get_cellchat_db):
        """Test the plot_modulators method."""
        # Mock the CellChat database
        mock_df_ligrec = pd.DataFrame({
            'ligand': ['ligand1', 'ligand2'],
            'receptor': ['receptor1', 'receptor2'],
            'pathway': ['pathway1', 'pathway2'],
            'signaling': ['Secreted Signaling', 'Cell-Cell Contact']
        })
        mock_get_cellchat_db.return_value = mock_df_ligrec
        
        # Mock the nichenet_lt dataframe
        mock_nichenet_lt = pd.DataFrame(
            np.random.rand(3, 2),
            index=['TF1', 'TF2', 'TF3'],
            columns=['ligand1', 'ligand2']
        )
        
        # Create a custom estimator class for testing
        class TestEstimator(SpatialCellularProgramsEstimator):
            def plot_modulators(self, use_expression=True, figsize=(10, 10), save_path=None):
                # Mock implementation
                return

        # Patch the pd.read_parquet to return our mock dataframe
        with patch('pandas.read_parquet', return_value=mock_nichenet_lt):
            # Create the test estimator
            estimator = TestEstimator(
                adata=self.adata,
                target_gene='target_gene',
                cluster_annot='rctd_cluster',
                layer='imputed_count',
                radius=200,
                contact_distance=50,
                tf_ligand_cutoff=0.01,
                colinks_path='dummy_path'
            )
            
            # Set necessary attributes for plot_modulators
            estimator.regulators = ['TF1', 'TF2', 'TF3']
            estimator.ligands = ['ligand1', 'ligand2']
            estimator.receptors = ['receptor1', 'receptor2']
            estimator.tfl_ligands = ['ligand1']
            estimator.tfl_regulators = ['TF1']
            estimator.lr_pairs = [('ligand1', 'receptor1'), ('ligand2', 'receptor2')]
            estimator.tfl_pairs = [('TF1', 'ligand1')]
            estimator.modulators = estimator.regulators + estimator.lr_pairs + estimator.tfl_pairs
            
            # Mock the model to return betas
            estimator.model = MagicMock()
            estimator.model.get_betas.return_value = torch.tensor([0.5, 0.3, 0.2, 0.1, 0.1])
            
            # Call plot_modulators
            estimator.plot_modulators(use_expression=True)
            
            # No need to check anything since we're using a mock implementation


if __name__ == '__main__':
    pytest.main(['-xvs', __file__])



import pyro
import torch.nn.functional as F
from torch.nn.modules.conv import _ConvNd
from torch.nn.modules.utils import _pair
from torch.nn.parameter import Parameter
import torch.nn as nn
import numpy as np
import torch    
import functools
from torch.nn.utils.parametrizations import weight_norm
from sklearn.linear_model import ARDRegression

from SpaceTravLR.models.vit_blocks import ViTBlock, get_positional_embeddings, patchify

pyro.clear_param_store()

device = torch.device(
    "mps" if torch.backends.mps.is_available() 
    else "cuda" if torch.cuda.is_available() 
    else "cpu"
)

use_conditional_conv = device == 'gpu'

class _cluster_routing(nn.Module):

    def __init__(self, num_clusters, pool_emb, num_experts, dropout_rate=0.1):
        super(_cluster_routing, self).__init__()
        
        self.dropout = nn.Dropout(dropout_rate)
        self.cluster_emb = nn.Embedding(num_clusters, pool_emb)
        self.fc = nn.Linear(pool_emb*2, num_experts)

    def forward(self, spatial_f, labels):
        spatial_f = spatial_f.flatten(1)
        emb = self.cluster_emb(labels)
        x = self.fc(torch.cat([spatial_f, emb], dim=1))
        x = self.dropout(x)
        return F.sigmoid(x)

    # def forward_(self, spatial_f, labels):
    #     spatial_f = spatial_f.flatten()
    #     emb = self.cluster_emb(labels)
    #     x = self.fc(torch.cat([spatial_f, emb]))
    #     x = self.dropout(x)
    #     return F.sigmoid(x)
    
class ConditionalConv2D(_ConvNd):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1,
                 padding=0, dilation=1, groups=1,
                 bias=True, padding_mode='zeros', num_experts=5, dropout_rate=0.1):
        kernel_size = _pair(kernel_size)
        stride = _pair(stride)
        padding = _pair(padding)
        dilation = _pair(dilation)
        super(ConditionalConv2D, self).__init__(
            in_channels, out_channels, kernel_size, stride, padding, dilation,
            False, _pair(0), groups, bias, padding_mode)

        self._avg_pooling = functools.partial(F.adaptive_avg_pool2d, output_size=(2, 2))
        pool_emb = torch.mul(*self._avg_pooling.keywords['output_size']) * in_channels
        
        self._routing_fn = _cluster_routing(
            num_clusters=in_channels,
            pool_emb=pool_emb,
            num_experts=num_experts, 
            dropout_rate=dropout_rate
        )

        self.weight = Parameter(torch.Tensor(
            num_experts, out_channels, in_channels // groups, *kernel_size))
        
        self.reset_parameters()

    def _conv_forward(self, input, weight):
        if self.padding_mode != 'zeros':
            return F.conv2d(F.pad(input, self._reversed_padding_repeated_twice, mode=self.padding_mode),
                            weight, self.bias, self.stride,
                            _pair(0), self.dilation, self.groups)
        return F.conv2d(input, weight, self.bias, self.stride,
                        self.padding, self.dilation, self.groups)
    

    def forward(self, inputs, input_labels):
        res = []
        
        assert inputs.shape[0] == input_labels.shape[0]
        pooled_inputs = self._avg_pooling(inputs)
        routing_weights = self._routing_fn(pooled_inputs, input_labels)
        
        # Get the index of the highest weight for each input
        max_weight_indices = torch.argmax(routing_weights, dim=1)
        
        # Select the kernel with the highest weight for each input
        kernels = self.weight[max_weight_indices]
        for inputx, kernel in zip(inputs, kernels):
            out = self._conv_forward(inputx.unsqueeze(0), kernel)
            res.append(out)


        # for inputx, label in zip(inputs, input_labels):
        #     inputx = inputx.unsqueeze(0)
        #     pooled_inputs = self._avg_pooling(inputx)
        #     routing_weights = self._routing_fn(pooled_inputs, label)
        #     kernels = torch.sum(routing_weights[:, None, None, None, None] * self.weight, 0)
        #     out = self._conv_forward(inputx, kernels)
        #     res.append(out)
        
        return torch.cat(res, dim=0)
    
class NicheAttentionNetwork(nn.Module):
     
    def __init__(self, n_regulators, in_channels, spatial_dim):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = in_channels
        self.spatial_dim = spatial_dim
        self.dim = n_regulators+1
        
        # if use_conditional_conv:
        #     self.conditional_conv = ConditionalConv2D(
        #         self.in_channels, self.in_channels, 1, num_experts=self.in_channels)
        
        self.conditional_conv = nn.Conv2d(self.in_channels, self.in_channels, 1)

        self.sigmoid = nn.Sigmoid()

        self.conv_layers = nn.Sequential(
            weight_norm(nn.Conv2d(in_channels, 32, kernel_size=3, padding='same')),
            nn.PReLU(init=0.1),
            nn.MaxPool2d(kernel_size=2, stride=2),
            
            weight_norm(nn.Conv2d(32, 64, kernel_size=3, padding='same')),
            nn.PReLU(init=0.1),
            nn.MaxPool2d(kernel_size=2, stride=2),

            weight_norm(nn.Conv2d(64, 128, kernel_size=3, padding='same')),
            nn.PReLU(init=0.1),
            nn.MaxPool2d(kernel_size=2, stride=2),
            
            nn.AdaptiveAvgPool2d(1),
            nn.Flatten()
        )

        # self.cluster_emb = nn.Embedding(self.in_channels, 128)

        self.mlp = nn.Sequential(
            nn.Linear(128, 64),
            nn.PReLU(init=0.1),
            nn.Linear(64, self.dim)
        )

        # self.alpha = nn.Parameter(torch.tensor(1.0), requires_grad=True)

        self.output_activation = nn.Tanh()


    def forward(self, spatial_maps, cluster_info):
        # att = self.sigmoid(self.conditional_conv(spatial_maps, cluster_info))
        att = self.sigmoid(self.conditional_conv(spatial_maps))
        out = att * spatial_maps
        out = self.conv_layers(out)
        # emb = self.cluster_emb(cluster_info) * self.alpha
        # out = out + emb 

        betas = self.mlp(out)
        betas = self.output_activation(betas)

        return betas
    
##Live Model
class CellularNicheNetwork(nn.Module):

    @staticmethod
    def make_vision_model(input_channels=1, out_dim=64, kernel_size=3):

        return nn.Sequential(
            weight_norm(nn.Conv2d(input_channels, 16, kernel_size=kernel_size, padding='same')),
            nn.BatchNorm2d(16),
            nn.PReLU(init=0.1),
            nn.MaxPool2d(kernel_size=2, stride=2),
            
            weight_norm(nn.Conv2d(16, 32, kernel_size=kernel_size, padding='same')),
            nn.BatchNorm2d(32),
            nn.PReLU(init=0.1),
            nn.MaxPool2d(kernel_size=2, stride=2),

            weight_norm(nn.Conv2d(32, out_dim, kernel_size=kernel_size, padding='same')),
            nn.BatchNorm2d(out_dim),
            nn.PReLU(init=0.1),
            nn.MaxPool2d(kernel_size=2, stride=2),
            
            nn.AdaptiveAvgPool2d(1),
            nn.Flatten()
        )
    
    @classmethod
    def from_pretrained(cls, trained_model, n_modulators, anchors=None, spatial_dim=64, n_clusters=7):
        cnn = cls.make_vision_model()
        cnn.load_state_dict(trained_model.conv_layers.state_dict())
        model = cls(n_modulators, anchors, spatial_dim, n_clusters)
        model.conv_layers = cnn
        return model

     
    def __init__(self, n_modulators, anchors=None, spatial_dim=64, n_clusters=7):
        super().__init__()
        self.in_channels = 1
        self.out_channels = 1
        self.spatial_dim = spatial_dim
        self.dim = n_modulators+1
        if anchors is None:
            anchors = np.ones(self.dim)

        self.anchors = torch.from_numpy(anchors).float().to(device)

        # self.anchors = torch.nn.Parameter(self.anchors, requires_grad=True)
        # self.conditional_conv = nn.Conv2d(self.in_channels, self.in_channels, 1)
        # self.sigmoid = nn.Sigmoid()

        self.conv_layers = self.make_vision_model(input_channels=self.in_channels)

        self.spatial_features_mlp = nn.Sequential(
            nn.Linear(n_clusters, 16),
            nn.PReLU(init=0.1),
            nn.Linear(16, 32),
            nn.PReLU(init=0.1),
            nn.Linear(32, 64)
        )

        self.mlp = nn.Sequential(
            nn.Linear(64, 64),
            nn.PReLU(init=0.1),
            nn.Linear(64, self.dim)
        )

        # self.output_activation = nn.Tanh()
        self.output_activation = nn.Sigmoid()
        # self.output_activation = nn.GELU()
        # self.output_activation = nn.Identity()
        # self.output_activation = nn.Softplus()


    def get_betas(self, spatial_maps, spatial_features):
        out = self.conv_layers(spatial_maps)
        sp_out = self.spatial_features_mlp(spatial_features)
        out = out+sp_out
        betas = self.mlp(out)
        betas = self.output_activation(betas)

        return betas*self.anchors
    
    @staticmethod
    def predict_y(inputs_x, betas):
        return torch.matmul(
                inputs_x.unsqueeze(1), 
                betas[:, 1:].unsqueeze(2)
            ).squeeze(1).squeeze(1) + \
                betas[:, 0]
    
    def forward(self, spatial_maps, inputs_x, spatial_features):
        betas = self.get_betas(spatial_maps, spatial_features)
        y_pred = self.predict_y(inputs_x, betas)
        
        return y_pred
    
class CellularViT(nn.Module):
    
    @classmethod
    def from_pretrained(cls, trained_model, n_modulators, anchors=None, spatial_dim=64, n_clusters=7):
        cnn = cls.make_vision_model()
        cnn.load_state_dict(trained_model.conv_layers.state_dict())
        model = cls(n_modulators, anchors, spatial_dim, n_clusters)
        model.conv_layers = cnn
        return model

     
    def __init__(self, n_modulators, anchors=None, spatial_dim=64, n_clusters=7, n_patches=4, n_blocks=4, hidden_d=16, n_heads=8):
        super().__init__()
        in_channels = 1
        self.out_channels = 1
        self.spatial_dim = spatial_dim
        self.dim = n_modulators+1
        if anchors is None:
            anchors = np.ones(self.dim)

        self.anchors = torch.from_numpy(anchors).float().to(device)
        
        self.dim = n_modulators+1
        if anchors is None:
            anchors = np.ones(self.dim)

        self.anchors = torch.from_numpy(anchors).float().to(device)
        
        self.in_channels = in_channels
        self.spatial_dim = spatial_dim
        
        chw = (in_channels, spatial_dim, spatial_dim) # ( C , H , W )
        self.n_patches = n_patches
        self.n_blocks = n_blocks
        self.n_heads = n_heads
        self.hidden_d = hidden_d
        
        # Input and patches sizes
        assert chw[1] % n_patches == 0, "Input shape not entirely divisible by number of patches"
        assert chw[2] % n_patches == 0, "Input shape not entirely divisible by number of patches"
        
        self.patch_size = (chw[1] / n_patches, chw[2] / n_patches)

        self.input_d = int(chw[0] * self.patch_size[0] * self.patch_size[1])
        self.linear_mapper = nn.Linear(self.input_d, self.hidden_d)
        
        self.class_token = nn.Parameter(torch.rand(1, self.hidden_d))       # Consider removing
        self.pos_embed = nn.Parameter(get_positional_embeddings(self.n_patches ** 2 + 1, self.hidden_d))
        self.pos_embed.requires_grad = False

        self.blocks = nn.ModuleList(
            [ViTBlock(hidden_d, n_heads) for _ in range(n_blocks)])


        self.spatial_features_mlp = nn.Sequential(
            nn.Linear(n_clusters, 16),
            nn.PReLU(init=0.1),
            nn.Linear(16, 32),
            nn.PReLU(init=0.1),
            nn.Linear(32, self.hidden_d)
        )


        self.mlp = nn.Sequential(
            nn.Linear(self.hidden_d, 32),
            nn.GELU(),
            nn.Dropout(0.1),
            nn.Linear(32, 16),
            nn.GELU(),
            nn.Dropout(0.1),
            nn.Linear(16, self.dim)
        )

        self.output_activation = nn.Sigmoid()
    
    
    def get_betas(self, spatial_maps, spatial_features):
        n, c, h, w = spatial_maps.shape 
        patches = patchify(spatial_maps, self.n_patches).to(self.pos_embed.device)
        sp_out = self.spatial_features_mlp(spatial_features)
        tokens = self.linear_mapper(patches) 
        tokens = torch.cat((self.class_token.expand(n, 1, -1), tokens), dim=1)
        out = tokens + self.pos_embed.repeat(n, 1, 1)
        
        for j, block in enumerate(self.blocks):
            out = block(out)
            
            
        betas = self.mlp(out[:, 0]+sp_out)
        betas = self.output_activation(betas) * 1.5
        
        return betas*self.anchors
    
    @staticmethod
    def predict_y(inputs_x, betas):
        return torch.matmul(
                inputs_x.unsqueeze(1), 
                betas[:, 1:].unsqueeze(2)
            ).squeeze(1).squeeze(1) + \
                betas[:, 0]
                
    
    def forward(self, spatial_maps, inputs_x, spatial_features):
        betas = self.get_betas(spatial_maps, spatial_features)
        y_pred = self.predict_y(inputs_x, betas)
        
        return y_pred


import pandas as pd 
import numpy as np
import matplotlib.pyplot as plt
from pysankey import sankey

from .shift import estimate_transition_probabilities


def get_macrostates(adata, delta_X, embedding, annot, n_neighbors=200, nn_transitions=10):
        transitions = estimate_transition_probabilities(
            adata, 
            delta_X,
            embedding=embedding,
            n_neighbors=n_neighbors,
            annot=annot
        )

        # Get the top nn_transitions transitions
        mask = np.argpartition(transitions, -nn_transitions, axis=1)[:, -nn_transitions:]
        row_indices = np.arange(transitions.shape[0])[:, None]

        transitions = transitions[row_indices, mask]
        idx2ct = {i: ct for i, ct in enumerate(adata.obs[annot])}
        ct_labels = pd.DataFrame(mask).replace(idx2ct)
        
        sankey_data = {'source': [], 'target': [], 'value': []}


        for target_ct in adata.obs[annot].unique():
            ct_mask = np.where(ct_labels == target_ct, 1, 0)
            target_transitions = transitions * ct_mask

            for source_ct in adata.obs[annot].unique():
                ct_transitions = target_transitions[adata.obs[annot] == source_ct]
                sankey_data['source'].append(source_ct)
                sankey_data['target'].append(target_ct)
                sankey_data['value'].append(ct_transitions.sum())

        sankey_df = pd.DataFrame(sankey_data)
        return sankey_df


def plot_pysankey(sankey_df, ax=None):
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(10, 10))
    sankey(
        left=sankey_df['source'], 
        right=sankey_df['target'], 
        leftWeight=sankey_df['value'], 
        rightWeight=sankey_df['value'],
        aspect=20,
        fontsize=12,
        ax=ax
    )
    return ax

